@@ -1,7 +1,7 @@
 /*
  *	archive.c
  *
- *	Archive File Access Module
+ *	archive file accessing module
  */
 
 #include "fd.h"
@@ -25,8 +25,6 @@ extern int tradlayout;
 #endif
 extern int sizeinfo;
 extern off_t marksize;
-extern off_t totalsize;
-extern off_t blocksize;
 extern namelist filestack[];
 extern char fullpath[];
 extern char typesymlist[];
@@ -38,7 +36,6 @@ extern int win_x;
 extern int win_y;
 extern char *deftmpdir;
 extern int hideclock;
-extern u_short today[];
 extern int n_args;
 
 #if	FD >= 2
@@ -81,6 +78,7 @@ static char *form_tar[] = {
 	"%a %l %u %g %s %m %d %{yt} %*f",	/* pax */
 	"%10a %u/%g %s %m %d %t %y %*f",	/* tar (UXP/DS) */
 	"%9a %u/%g %s %m %d %t %y %*f",		/* traditional */
+	"%a %u %g %s %m %d %t %y %*f",		/* AIX */
 	"%a %u/%g %m %d %t %y %*f",		/* IRIX */
 #endif	/* !MSDOS */
 	NULL
@@ -149,9 +147,6 @@ static char *form_tar[] = {
 #endif	/* FD < 2 */
 
 #define	iswhitespace(c)	((c) == ' ' || (c) == '\t')
-#define	S_IREAD_ALL	(S_IRUSR | S_IRGRP | S_IROTH)
-#define	S_IWRITE_ALL	(S_IWUSR | S_IWGRP | S_IWOTH)
-#define	S_IEXEC_ALL	(S_IXUSR | S_IXGRP | S_IXOTH)
 
 #ifndef	_NOBROWSE
 static VOID NEAR copyargvar __P_((int, char **));
@@ -190,7 +185,8 @@ static int NEAR parsearchive __P_((FILE *, launchtable *, namelist *, int *));
 static VOID NEAR unpackerror __P_((VOID_A));
 static int NEAR readarchive __P_((char *, launchtable *, int));
 static char *NEAR searcharcdir __P_((char *, int));
-static int NEAR undertmp(char *);
+static char *NEAR archoutdir __P_((VOID_A));
+static int NEAR undertmp __P_((char *));
 #ifdef	_NODOSDRIVE
 static char *NEAR genfullpath __P_((char *, char *, char *));
 #else
@@ -261,7 +257,8 @@ char **browsevar = NULL;
 #if	FD >= 2
 static char *autoformat[] = {
 # if	MSDOS
-	"%*f\n%s %x %x %y-%m-%d %t %a",		/* LHa (-v) */
+	"%*f\n%s %x %x %y-%m-%d %t %a",		/* LHa (v) */
+	" %*f\n%s %x %x %d-%m-%y %t %a",	/* RAR (v) */
 	"%a %u/%g %s %m %d %t %y %*f",		/* tar (traditional) */
 	"%a %u/%g %s %y-%m-%d %t %*f",		/* tar (GNU >=1.12) */
 	"%a %l %u %g %s %m %d %{yt} %*f",	/* ls or pax */
@@ -269,8 +266,10 @@ static char *autoformat[] = {
 	" %s %y-%m-%d %t %*f",			/* zip */
 	" %s %x %x %x %y-%m-%d %t %*f",		/* pkunzip */
 	" %s %x %x %d %m %y %t %*f",		/* zoo */
-	"%1x %12f %s %x %x %y-%m-%d %t %a",	/* LHa (-l) */
+	"%1x %12f %s %x %x %y-%m-%d %t %a",	/* LHa (l) */
+	" %f %s %x %x %d-%m-%y %t %a",		/* RAR (l) */
 # else	/* !MSDOS */
+	" %*f\n%s %x %x %d-%m-%y %t %a",	/* RAR (v) */
 	"%a %u/%g %s %m %d %t %y %*f",		/* tar (SVR4) */
 	"%a %u/%g %s %y-%m-%d %t %*f",		/* tar (GNU >=1.12) */
 	"%a %l %u %g %s %m %d %{yt} %*f",	/* ls or pax */
@@ -281,9 +280,10 @@ static char *autoformat[] = {
 	"%9a %u/%g %s %m %d %t %y %*f",		/* tar (traditional) */
 	"%9a %u/%g %s %x %m %d %{yt} %*f",	/* LHa */
 	"%a %u/%g %m %d %t %y %*f",		/* tar (IRIX) */
+	" %f %s %x %x %d-%m-%y %t %a",		/* RAR (l) */
 # endif	/* !MSDOS */
 };
-#define	AUTOFORMATSIZ	((int)(sizeof(autoformat) / sizeof(char *)))
+#define	AUTOFORMATSIZ	arraysize(autoformat)
 #endif	/* FD >= 2 */
 
 
@@ -459,9 +459,7 @@ char *buf;
 	int i, c, len;
 	u_int mode;
 
-	if (!*buf) return(0);
 	len = strlen(buf);
-
 	if (len < 9) {
 		mode = (tmp -> st_mode & S_IFMT) | S_IREAD_ALL | S_IWRITE_ALL;
 		while (--len >= 0) {
@@ -485,6 +483,7 @@ char *buf;
 					break;
 				case 'w':
 				case '-':
+				case '.':
 				case ' ':
 					break;
 				case 'v':
@@ -547,9 +546,9 @@ char *buf;
 			mode |= typelist[i];
 		}
 	}
-
 	tmp -> st_mode = mode;
-	return(1);
+
+	return(len);
 }
 
 #if	FD >= 2
@@ -571,6 +570,7 @@ int *scorep;
 			*scorep += 4;
 		}
 	}
+
 	return(strndup2(s, len));
 }
 
@@ -584,6 +584,7 @@ int len;
 	cp = strndup2(s, len);
 	tmp = newkanjiconv(cp, fnamekcode, DEFCODE, L_FNAME);
 	if (tmp != cp) free(cp);
+
 	return(tmp);
 }
 # endif	/* !_NOKANJIFCONV */
@@ -599,7 +600,8 @@ int skip;
 	uid_t uid;
 	gid_t gid;
 # endif
-	struct tm tm;
+	time_t now;
+	struct tm tm, *tp;
 	off_t n;
 	int i, ch, l, len, hit, err, err2, score;
 	char *cp, *s, *buf, *rawbuf;
@@ -678,7 +680,8 @@ int skip;
 
 		if (len < 0) {
 			if (!iswhitespace(*form)
-			&& (*form != '%' || *(form + 1) == '%')) ch = *form;
+			&& (*form != '%' || *(form + 1) == '%'))
+				ch = *form;
 			for (len = 0; line[len]; len++) {
 				if (ch) {
 					if (ch == line[len]) break;
@@ -701,7 +704,7 @@ int skip;
 				break;
 			case 'u':
 # ifndef	NOUID
-				cp = sscanf2(buf, "%-*d%$",
+				cp = sscanf2(buf, "%-<*d%$",
 					sizeof(uid_t), &uid);
 				if (cp) tmp -> st_uid = uid;
 				else if ((up = finduid(0, buf)))
@@ -711,7 +714,7 @@ int skip;
 				break;
 			case 'g':
 # ifndef	NOUID
-				cp = sscanf2(buf, "%-*d%$",
+				cp = sscanf2(buf, "%-<*d%$",
 					sizeof(gid_t), &gid);
 				if (cp) tmp -> st_gid = gid;
 				else if ((gp = findgid(0, buf)))
@@ -811,7 +814,9 @@ int skip;
 				if (isdelim(rawbuf, i - 1)) {
 					tmp -> st_mode &= ~S_IFMT;
 					tmp -> st_mode |= S_IFDIR;
-					if (i > 1) i--;
+					while (i > 1)
+						if (!isdelim(rawbuf, --i - 1))
+							break;
 				}
 				if (tmp -> name) free(tmp -> name);
 				tmp -> name = readfname(rawbuf, i);
@@ -838,21 +843,24 @@ int skip;
 				err = 0;
 				break;
 			default:
-				free(buf);
-				free(rawbuf);
-				if (tmp -> name) free(tmp -> name);
-# ifndef	NOSYMLINK
-				if (tmp -> linkname) free(tmp -> linkname);
-# endif
-				return(-1);
-/*NOTREACHED*/
+				hit = -1;
 				break;
 		}
+
+		free(buf);
+		free(rawbuf);
+
+		if (hit < 0) {
+			if (tmp -> name) free(tmp -> name);
+# ifndef	NOSYMLINK
+			if (tmp -> linkname) free(tmp -> linkname);
+# endif
+			return(hit);
+		}
+
 		if (!hit) score += 5;
 		else if (hit <= err2) score += err2;
 		score += err;
-		free(buf);
-		free(rawbuf);
 		line += len;
 		if (!ch) while (iswhitespace(*line)) line++;
 	}
@@ -866,13 +874,25 @@ int skip;
 	}
 
 	if (!(tmp -> st_mode & S_IFMT)) tmp -> st_mode |= S_IFREG;
-	if ((tmp -> st_mode & S_IFMT) == S_IFDIR) tmp -> flags |= F_ISDIR;
-	else if ((tmp -> st_mode & S_IFMT) == S_IFLNK) tmp -> flags |= F_ISLNK;
+	if (s_isdir(tmp)) tmp -> flags |= F_ISDIR;
+	else if (s_islnk(tmp)) tmp -> flags |= F_ISLNK;
 	if (tm.tm_year < 0) {
-		tm.tm_year = today[0];
+		now = time2();
+# ifdef	DEBUG
+		_mtrace_file = "localtime(start)";
+		tp = localtime(&now);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "localtime(end)";
+			malloc(0);	/* dummy malloc */
+		}
+# else
+		tp = localtime(&now);
+# endif
+		tm.tm_year = tp -> tm_year;
 		if (tm.tm_mon < 0 || tm.tm_mday < 0) tm.tm_year = 1970 - 1900;
-		else if (tm.tm_mon > today[1]
-		|| (tm.tm_mon == today[1] && tm.tm_mday > today[2]))
+		else if (tm.tm_mon > tp -> tm_mon
+		|| (tm.tm_mon == tp -> tm_mon && tm.tm_mday > tp -> tm_mday))
 			tm.tm_year--;
 	}
 	else {
@@ -929,13 +949,15 @@ int field, *eolp;
 			if (eolp) {
 				for (j = i; line[j]; j++) {
 					if ((sp < 255 && j >= sp)
-					|| iswhitespace(line[j])) break;
+					|| iswhitespace(line[j]))
+						break;
 				}
 				*eolp = j;
 			}
 			return(i);
 		}
 	}
+
 	return((field < 0) ? f : -1);
 }
 
@@ -971,8 +993,8 @@ int no;
 		else i = &(line[eol]) - cp;
 	}
 	if (!i || &(cp[i]) > &(line[eol])) i = &(line[eol]) - cp;
-
 	strncpy2(buf, cp, i);
+
 	return(buf);
 }
 
@@ -988,7 +1010,8 @@ int max;
 	uid_t uid;
 	gid_t gid;
 # endif
-	struct tm tm;
+	time_t now;
+	struct tm tm, *tp;
 	off_t n;
 	int i, skip;
 	char *cp, *buf;
@@ -1026,17 +1049,17 @@ int max;
 	tmp -> name = strdup2(buf);
 
 	getfield(buf, line, skip, list, F_MODE);
-	if (!readattr(tmp, buf)) tmp -> st_mode = 0644;
+	readattr(tmp, buf);
 	if (!(tmp -> st_mode & S_IFMT)) tmp -> st_mode |= S_IFREG;
-	if ((tmp -> st_mode & S_IFMT) == S_IFDIR) tmp -> flags |= F_ISDIR;
-	else if ((tmp -> st_mode & S_IFMT) == S_IFLNK) tmp -> flags |= F_ISLNK;
+	if (s_isdir(tmp)) tmp -> flags |= F_ISDIR;
+	else if (s_islnk(tmp)) tmp -> flags |= F_ISLNK;
 
 # ifndef	NOUID
 	getfield(buf, line, skip, list, F_UID);
-	if (sscanf2(buf, "%-*d%$", sizeof(uid_t), &uid)) tmp -> st_uid = uid;
+	if (sscanf2(buf, "%-<*d%$", sizeof(uid_t), &uid)) tmp -> st_uid = uid;
 	else tmp -> st_uid = ((up = finduid(0, buf))) ? up -> uid : (uid_t)-1;
 	getfield(buf, line, skip, list, F_GID);
-	if (sscanf2(buf, "%-*d%$", sizeof(gid_t), &gid)) tmp -> st_gid = gid;
+	if (sscanf2(buf, "%-<*d%$", sizeof(gid_t), &gid)) tmp -> st_gid = gid;
 	else tmp -> st_gid = ((gp = findgid(0, buf))) ? gp -> gid : (gid_t)-1;
 # endif
 	getfield(buf, line, skip, list, F_SIZE);
@@ -1064,9 +1087,21 @@ int max;
 	else if ((i = atoi2(buf)) >= 0) tm.tm_year = i;
 	else if (list -> field[F_YEAR] == list -> field[F_TIME]
 	&& strchr(buf, ':')) {
-		tm.tm_year = today[0];
-		if (tm.tm_mon > today[1]
-		|| (tm.tm_mon == today[1] && tm.tm_mday > today[2]))
+		now = time2();
+# ifdef	DEBUG
+		_mtrace_file = "localtime(start)";
+		tp = localtime(&now);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "localtime(end)";
+			malloc(0);	/* dummy malloc */
+		}
+# else
+		tp = localtime(&now);
+# endif
+		tm.tm_year = tp -> tm_year;
+		if (tm.tm_mon > tp -> tm_mon
+		|| (tm.tm_mon == tp -> tm_mon && tm.tm_mday > tp -> tm_mday))
 			tm.tm_year--;
 	}
 	else tm.tm_year = 1970;
@@ -1095,8 +1130,8 @@ int max;
 # else
 		logical_access(tmp -> st_mode, tmp -> st_uid, tmp -> st_gid);
 # endif
-
 	free(buf);
+
 	return(0);
 }
 #endif	/* FD < 2 */
@@ -1111,7 +1146,7 @@ char *file, *dir;
 	if (browselist) {
 		int i;
 
-		if (!browsevar) arch = strdup2("");
+		if (!browsevar) arch = strdup2(nullstr);
 		else {
 			len = 0;
 			for (i = 0; browsevar[i]; i++)
@@ -1136,59 +1171,59 @@ char *file, *dir;
 
 #ifndef	_NOTRADLAYOUT
 	if (istradlayout()) {
-		locate(0, TL_PATH);
-		putterm(l_clear);
+		Xlocate(0, TL_PATH);
+		Xputterm(L_CLEAR);
 
-		locate(TC_PATH, TL_PATH);
-		putterm(t_standout);
+		Xlocate(TC_PATH, TL_PATH);
+		Xputterm(T_STANDOUT);
 # ifndef	_NOBROWSE
 		if (browselist) {
-			cputs2(TS_BROWSE);
+			Xcputs2(TS_BROWSE);
 			len = TD_BROWSE;
 		}
 		else
 # endif
 		{
-			cputs2(TS_ARCH);
+			Xcputs2(TS_ARCH);
 			len = TD_ARCH;
 		}
-		putterm(end_standout);
-		cprintf2("%-*.*k", len, len, arch);
+		Xputterm(END_STANDOUT);
+		cputstr(len, arch);
 		free(arch);
 
-		locate(TC_MARK, TL_PATH);
-		cprintf2("%<*d", TD_MARK, mark);
-		putterm(t_standout);
-		cputs2(TS_MARK);
-		putterm(end_standout);
-		cprintf2("%<'*qd", TD_SIZE, marksize);
+		Xlocate(TC_MARK, TL_PATH);
+		Xcprintf2("%<*d", TD_MARK, mark);
+		Xputterm(T_STANDOUT);
+		Xcputs2(TS_MARK);
+		Xputterm(END_STANDOUT);
+		Xcprintf2("%<'*qd", TD_SIZE, marksize);
 
-		tflush();
+		Xtflush();
 		return;
 	}
 #endif	/* !_NOTRADLAYOUT */
 
-	locate(0, L_PATH);
-	putterm(l_clear);
+	Xlocate(0, L_PATH);
+	Xputterm(L_CLEAR);
 
-	locate(C_PATH, L_PATH);
-	putterm(t_standout);
+	Xlocate(C_PATH, L_PATH);
+	Xputterm(T_STANDOUT);
 #ifndef	_NOBROWSE
 	if (browselist) {
-		cputs2(S_BROWSE);
+		Xcputs2(S_BROWSE);
 		len = D_BROWSE;
 	}
 	else
 #endif
 	{
-		cputs2(S_ARCH);
+		Xcputs2(S_ARCH);
 		len = D_ARCH;
 	}
-	putterm(end_standout);
-	cprintf2("%-*.*k", len, len, arch);
+	Xputterm(END_STANDOUT);
+	cputstr(len, arch);
 	free(arch);
 
-	tflush();
+	Xtflush();
 }
 
 static int NEAR dircmp(s1, s2)
@@ -1208,6 +1243,7 @@ char *s1, *s2;
 		while (s1[i + 1] == _SC_) i++;
 		while (s2[j + 1] == _SC_) j++;
 	}
+
 	return(s1[i]);
 }
 
@@ -1237,21 +1273,21 @@ char *s1, *s2;
 		return((i == j) ? 1 : 0);
 	}
 	if (i && s2[j]) while (s2[j + 1] == _SC_) j++;
+
 	return(j);
 }
 
 static char *NEAR pseudodir(namep)
 namelist *namep;
 {
-	char *cp, *tmp;
+	char *cp, *next;
 	int i, len;
 	u_short ent;
 
-	cp = namep -> name;
-	while ((tmp = strdelim(cp, 0))) {
-		while (*(tmp + 1) == _SC_) tmp++;
-		if (!*(tmp + 1)) break;
-		len = tmp - (namep -> name);
+	for (cp = namep -> name; (next = strdelim(cp, 0)); cp = ++next) {
+		while (*(next + 1) == _SC_) next++;
+		if (!*(next + 1)) break;
+		len = next - namep -> name;
 		if (!len) len++;
 		for (i = 0; i < maxfile; i++) {
 			if (isdir(&(filelist[i]))
@@ -1263,21 +1299,22 @@ namelist *namep;
 			filelist[i].name = realloc2(filelist[i].name, len + 1);
 			strncpy2(filelist[i].name, namep -> name, len);
 		}
-		cp = tmp + 1;
 	}
+
 	if (isdir(namep) && !isdotdir(namep -> name))
 	for (i = 0; i < maxfile; i++) {
 		if (isdir(&(filelist[i]))
 		&& !dircmp(filelist[i].name, namep -> name)) {
-			tmp = filelist[i].name;
+			cp = filelist[i].name;
 			ent = filelist[i].ent;
 			memcpy((char *)&(filelist[i]), (char *)namep,
 				sizeof(namelist));
-			filelist[i].name = tmp;
+			filelist[i].name = cp;
 			filelist[i].ent = ent;
 			return(NULL);
 		}
 	}
+
 	return(namep -> name);
 }
 
@@ -1345,6 +1382,7 @@ char *s, **argv;
 # ifndef	PATHNOCASE
 	pathignorecase = duppathignorecase;
 # endif
+
 	return(ret);
 }
 #endif	/* FD >= 2 */
@@ -1563,6 +1601,7 @@ int *linenop;
 				(nline - needline + 1) * sizeof(char *));
 	}
 	*linenop += needline;
+
 #if	FD >= 2
 	if (form0) free(form0);
 	free(scorelist);
@@ -1579,17 +1618,17 @@ static VOID NEAR unpackerror(VOID_A)
 		fputnl(stderr);
 	}
 	else {
-		locate(0, n_line - 1);
-		cputnl();
+		Xlocate(0, n_line - 1);
+		Xcputnl();
 		hideclock = 1;
 		warning(0, UNPNG_K);
 	}
 }
 
-static int NEAR readarchive(file, list, argset)
+static int NEAR readarchive(file, list, flags)
 char *file;
 launchtable *list;
-int argset;
+int flags;
 {
 	namelist tmp;
 	FILE *fp;
@@ -1597,7 +1636,7 @@ int argset;
 	int i, no, max, wastty;
 
 	wastty = isttyiomode;
-	if (!(fp = popenmacro(list -> comm, file, argset))) return(-1);
+	if (!(fp = popenmacro(list -> comm, file, flags))) return(-1);
 	if (wastty) Xwaitmes();
 
 	for (i = 0; i < (int)(list -> topskip); i++) {
@@ -1607,11 +1646,11 @@ int argset;
 
 	maxfile = 0;
 	addlist();
-	filelist[0].name = strdup2("..");
+	filelist[0].name = strdup2(parentpath);
 #ifndef	NOSYMLINK
 	filelist[0].linkname = NULL;
 #endif
-	filelist[0].flags = F_ISDIR;
+	filelist[0].flags = (F_ISEXE | F_ISRED | F_ISWRI | F_ISDIR);
 	filelist[0].tmpflags = F_STAT;
 	maxfile++;
 #ifdef	HAVEFLAGS
@@ -1707,6 +1746,7 @@ int argset;
 		}
 		return(-1);
 	}
+
 	return(maxfile);
 }
 
@@ -1715,25 +1755,30 @@ reg_t *re;
 char *arcre;
 {
 	char *cp, *tmp;
-	int i, j, n, len;
+	int i, j, n, len, parent;
 
+	parent = (*archivedir) ? 0 : -1;
+	/* omit filelist[0] as pseudo ".." */
 	for (i = 1; i < maxarcf; i++) {
 		if (!*archivedir) len = 0;
 		else if (!(len = dirmatchlen(archivedir, arcflist[i].name)))
 			continue;
 
-		cp = arcflist[i].name + len;
-		if (len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
-		if (!arcflist[i].name[len]) {
-			memcpy((char *)&(filelist[0]),
-				(char *)&(arcflist[i]), sizeof(namelist));
-			filelist[0].name = arcflist[0].name;
-#ifndef	NOSYMLINK
-			filelist[0].linkname = NULL;
-#endif
+		cp = &(arcflist[i].name[len]);
+		if (*cp && len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
+		if (!*cp) {
+			parent = i;
 			continue;
 		}
-		if ((tmp = strdelim(cp, 0))) while (*(++tmp) == _SC_);
+
+		if (!(tmp = strdelim(cp, 0))) len = strlen(cp);
+		else {
+			len = (tmp == cp) ? 1 : tmp - cp;
+			while (*(++tmp) == _SC_);
+		}
+
+		if (parent <= 0 && len == 2 && cp[0] == '.' && cp[1] == '.')
+			parent = i;
 		if ((tmp && *tmp) || (re && !regexp_exec(re, cp, 1))) continue;
 		if (arcre) {
 			if (!(n = searcharc(arcre, arcflist, maxarcf, i)))
@@ -1751,10 +1796,20 @@ char *arcre;
 #ifndef	NOSYMLINK
 		filelist[maxfile].linkname = arcflist[i].linkname;
 #endif
-		if (isfile(&(filelist[maxfile])))
-			totalsize += getblock(filelist[maxfile].st_size);
 		maxfile++;
 	}
+
+	if (parent < 0) return;
+	for (i = 0; i < maxfile; i++)
+		if (isdotdir(filelist[i].name) == 1) return;
+	memmove((char *)&(filelist[1]), (char *)&(filelist[0]),
+		maxfile++ * sizeof(namelist));
+	memcpy((char *)&(filelist[0]),
+		(char *)&(arcflist[parent]), sizeof(namelist));
+	filelist[0].name = arcflist[0].name;
+#ifndef	NOSYMLINK
+	filelist[0].linkname = NULL;
+#endif
 }
 
 static char *NEAR searcharcdir(file, flen)
@@ -1765,14 +1820,14 @@ int flen;
 	int i, len;
 
 	errno = ENOENT;
-	for (i = 1; i < maxarcf; i++) {
+	for (i = 0; i < maxarcf; i++) {
 		if (!*archivedir) len = 0;
 		else if (!(len = dirmatchlen(archivedir, arcflist[i].name)))
 			continue;
 
-		cp = arcflist[i].name + len;
-		if (len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
-		if (!arcflist[i].name[len]) {
+		cp = &(arcflist[i].name[len]);
+		if (*cp && len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
+		if (!*cp) {
 			if (file) continue;
 		}
 		else {
@@ -1782,6 +1837,8 @@ int flen;
 				len = (tmp == cp) ? 1 : tmp - cp;
 				while (*(++tmp) == _SC_);
 			}
+			if (!*file && len == 2 && cp[0] == '.' && cp[1] == '.')
+				return(arcflist[0].name);
 			if ((tmp && *tmp) || len != flen
 			|| strnpathcmp(file, cp, flen))
 				continue;
@@ -1792,9 +1849,33 @@ int flen;
 		}
 		return(arcflist[i].name);
 	}
+
 	return(NULL);
 }
 
+static char *NEAR archoutdir(VOID_A)
+{
+	char *cp, *file;
+
+	if (!*archivedir) return((char *)-1);
+	if (!(file = searcharcdir(NULL, 0))) return(NULL);
+
+	cp = &(file[(int)strlen(file) - 1]);
+	while (cp > file && *cp == _SC_) cp--;
+#ifdef	BSPATHDELIM
+	if (onkanji1(file, cp - file)) cp++;
+#endif
+	cp = strrdelim2(file, cp);
+	if (!cp) *archivedir = '\0';
+	else {
+		if (cp == file) copyrootpath(archivedir);
+		else strncpy2(archivedir, file, cp - file);
+		file = ++cp;
+	}
+
+	return(file);
+}
+
 char *archchdir(path)
 char *path;
 {
@@ -1803,11 +1884,11 @@ char *path;
 
 	if (findpattern) free(findpattern);
 	findpattern = NULL;
-	if (!path || !*path) path = "..";
 #ifndef	_NOBROWSE
 	if (browselist) {
-		int i, n, dupfilepos;
+		int i, n, flags, dupfilepos;
 
+		if (!path || !*path) path = parentpath;
 		if ((cp = strdelim(path, 0))) {
 			for (i = 1; cp[i]; i++) if (cp[i] != _SC_) break;
 			if (cp[i]) {
@@ -1820,7 +1901,9 @@ char *path;
 
 		n = browselevel;
 		cp = browselist[n].ext;
+		flags = (F_NOCONFIRM | F_ARGSET | F_NOADDOPT | F_IGNORELIST);
 		dupfilepos = filepos;
+
 		if (path == filelist[filepos].name) i = filepos;
 		else for (i = 0; i < maxfile; i++)
 			if (!strcmp(path, filelist[i].name)) break;
@@ -1832,7 +1915,7 @@ char *path;
 
 		pushbrowsevar(path);
 		if (cp && !(browselist[n].flags & LF_DIRNOPREP))
-			execusercomm(cp, path, 1, 1, 1);
+			execusercomm(cp, path, flags);
 		if (isdotdir(path)) {
 			filepos = dupfilepos;
 			popbrowsevar();
@@ -1840,51 +1923,39 @@ char *path;
 		}
 
 		if (browselist[n + 1].comm
-		&& !(browselist[n].flags & LF_DIRLOOP)) n++;
-		if (dolaunch(&(browselist[n]), 1) < 0
+		&& !(browselist[n].flags & LF_DIRLOOP))
+			n++;
+		if (dolaunch(&(browselist[n]), flags) < 0
 		|| !isarchbr(&(browselist[n]))) {
 			popbrowsevar();
 			return(path);
 		}
 		browselevel = n;
-		return("..");
+		return(parentpath);
 	}
 #endif	/* !_NOBROWSE */
+
+	if (!path) return(archoutdir());
 	strcpy(duparcdir, archivedir);
 	do {
 		if (*path == _SC_) len = 1;
 		else if ((cp = strdelim(path, 0))) len = cp - path;
 		else len = strlen(path);
-		if (len != 2 || strncmp(path, "..", len)) {
-			if (!searcharcdir(path, len)) {
-				strcpy(archivedir, duparcdir);
-				errno = ENOENT;
-				return(NULL);
-			}
+
+		cp = path;
+		if (len == 2 && path[0] == '.' && path[1] == '.') cp = nullstr;
+		if (searcharcdir(cp, len)) {
 			if (*(cp = archivedir)) cp = strcatdelim(archivedir);
 			strncpy2(cp, path, len);
-			file = "..";
+			file = parentpath;
 		}
-		else if (!*archivedir) return((char *)-1);
-		else {
-			if (!(file = searcharcdir(NULL, 0))) {
-				strcpy(archivedir, duparcdir);
-				errno = ENOENT;
-				return(NULL);
-			}
-			cp = file + (int)strlen(file) - 1;
-			while (cp > file && *cp == _SC_) cp--;
-#ifdef	BSPATHDELIM
-			if (onkanji1(file, cp - file)) cp++;
-#endif
-			cp = strrdelim2(file, cp);
-			if (!cp) *archivedir = '\0';
-			else {
-				if (cp == file) strcpy(archivedir, _SS_);
-				else strncpy2(archivedir, file, cp - file);
-				file = cp + 1;
-			}
+		else if (*cp || !(file = archoutdir())) {
+			strcpy(archivedir, duparcdir);
+			errno = ENOENT;
+			return(NULL);
 		}
+		else if (file == (char *)-1) break;
+
 		path += len;
 		while (*path == _SC_) path++;
 	} while (*path);
@@ -1899,7 +1970,7 @@ int flen, argc;
 char ***argvp;
 {
 	char *cp, *tmp, *new, *file, dir[MAXPATHLEN], duparcdir[MAXPATHLEN];
-	int i, len;
+	int i, len, parent;
 
 # ifdef	_USEDOSPATH
 	if (_dospath(path)) return(argc);
@@ -1919,25 +1990,30 @@ char ***argvp;
 		flen -= ++file - path;
 	}
 
+	/* omit filelist[0] as pseudo ".." */
 	for (i = 1; i < maxarcf; i++) {
 		if (!*archivedir) len = 0;
 		else if (!(len = dirmatchlen(archivedir, arcflist[i].name)))
 			continue;
 
-		cp = arcflist[i].name + len;
-		if (len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
-		if (!arcflist[i].name[len]) continue;
+		cp = &(arcflist[i].name[len]);
+		if (*cp && len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
+		if (!*cp) continue;
 
 		if (!(tmp = strdelim(cp, 0))) len = strlen(cp);
 		else {
 			len = (tmp == cp) ? 1 : tmp - cp;
 			while (*(++tmp) == _SC_);
 		}
-		if ((tmp && *tmp) || strnpathcmp(file, cp, flen)) continue;
+
+		parent = 0;
+		if (len == 2 && cp[0] == '.' && cp[1] == '.') parent++;
+		else if (tmp && *tmp) continue;
+		if (strnpathcmp(file, cp, flen)) continue;
 
 		new = malloc2(len + 1 + 1);
 		strncpy(new, cp, len);
-		if (isdir(&(arcflist[i]))) new[len++] = _SC_;
+		if (parent || isdir(&(arcflist[i]))) new[len++] = _SC_;
 		new[len] = '\0';
 		if (finddupl(new, argc, *argvp)) {
 			free(new);
@@ -1949,13 +2025,14 @@ char ***argvp;
 		(*argvp)[argc++] = new;
 	}
 	strcpy(archivedir, duparcdir);
+
 	return(argc);
 }
 #endif	/* !_NOCOMPLETE */
 
-int dolaunch(list, argset)
+int dolaunch(list, flags)
 launchtable *list;
-int argset;
+int flags;
 {
 #ifndef	_NODOSDRIVE
 	int drive;
@@ -1968,24 +2045,23 @@ int argset;
 	drive = 0;
 #endif
 
-	if (argset) /*EMPTY*/;
+	if (flags & F_ARGSET) /*EMPTY*/;
 	else if (archivefile) {
 		if (!(tmpdir = tmpunpack(1))) {
-			putterm(t_bell);
+			Xputterm(T_BELL);
 			return(-1);
 		}
 	}
 #ifndef	_NODOSDRIVE
-	else if ((drive = tmpdosdupl("", &tmpdir, 1)) < 0) {
-		putterm(t_bell);
+	else if ((drive = tmpdosdupl(nullstr, &tmpdir, 1)) < 0) {
+		Xputterm(T_BELL);
 		return(-1);
 	}
 #endif
 
 	if (!isarchbr(list)) {
-		execusercomm(list -> comm, filelist[filepos].name,
-			1, argset, 1);
-		if (argset) /*EMPTY*/;
+		ptyusercomm(list -> comm, filelist[filepos].name, flags);
+		if (flags & F_ARGSET) /*EMPTY*/;
 #ifndef	_NODOSDRIVE
 		else if (drive) removetmp(tmpdir, filelist[filepos].name);
 #endif
@@ -1994,7 +2070,7 @@ int argset;
 	}
 
 	pusharchdupl();
-	if (argset) /*EMPTY*/;
+	if (flags & F_ARGSET) /*EMPTY*/;
 #ifndef	_NODOSDRIVE
 	else if (drive) {
 		archduplp -> v_fullpath = strdup2(fullpath);
@@ -2011,7 +2087,7 @@ int argset;
 	}
 	archivefile = (filelist && filepos < maxfile)
 		? filelist[filepos].name : NULL;
-	if (!archivefile) archivefile = "";
+	if (!archivefile) archivefile = nullstr;
 	archivefile = strdup2(archivefile);
 	*archivedir = '\0';
 	archtmpdir = tmpdir;
@@ -2027,7 +2103,7 @@ int argset;
 	maxent = 0;
 	sorton = 0;
 
-	if (readarchive(archivefile, launchp, argset) < 0) {
+	if (readarchive(archivefile, launchp, flags) < 0) {
 		arcflist = NULL;
 		poparchdupl();
 		return(-1);
@@ -2053,17 +2129,20 @@ int launcher(VOID_A)
 #ifndef	_NOBROWSE
 	if (browselist) {
 		char *cp;
+		int flags;
 
 		n = browselevel;
 		cp = browselist[n].ext;
+		flags = (F_NOCONFIRM | F_ARGSET | F_NOADDOPT | F_IGNORELIST);
 
 		pushbrowsevar(filelist[filepos].name);
 		if (cp && !(browselist[n].flags & LF_FILENOPREP))
-			execusercomm(cp, filelist[filepos].name, 1, 1, 1);
+			execusercomm(cp, filelist[filepos].name, flags);
 
 		if (browselist[n + 1].comm
-		&& !(browselist[n].flags & LF_FILELOOP)) n++;
-		if (dolaunch(&(browselist[n]), 1) < 0
+		&& !(browselist[n].flags & LF_FILELOOP))
+			n++;
+		if (dolaunch(&(browselist[n]), flags) < 0
 		|| !isarchbr(&(browselist[n]))) {
 			popbrowsevar();
 			return(-1);
@@ -2087,7 +2166,7 @@ int launcher(VOID_A)
 	}
 	if (i >= maxlaunch) return(0);
 
-	return(dolaunch(&(launchlist[i]), 0));
+	return(dolaunch(&(launchlist[i]), F_NOCONFIRM | F_IGNORELIST));
 }
 
 static int NEAR undertmp(path)
@@ -2116,12 +2195,19 @@ static char *NEAR genfullpath(path, file, full, tmpdir)
 char *path, *file, *full, *tmpdir;
 #endif
 {
+#if	MSDOS
+	int drive;
+#endif
 	char *cp;
 
 	cp = file;
 #ifdef	_USEDOSPATH
+# if	MSDOS
+	if ((drive = _dospath(cp))) cp += 2;
+# else
 	if (_dospath(cp)) cp += 2;
-#endif
+# endif
+#endif	/* _USEDOSPATH */
 
 #ifndef	_NODOSDRIVE
 	if (tmpdir) strcpy(path, tmpdir);
@@ -2136,21 +2222,16 @@ char *path, *file, *full, *tmpdir;
 		return(NULL);
 	}
 #if	MSDOS
-	else if (cp > file) {
-		int drive;
-
-		path[0] = *file;
-		path[1] = ':';
-		path[2] = _SC_;
-		if ((drive = toupper2(*file)) == toupper2(*full))
-			strcpy(&(path[3]), &(full[3]));
-		else if (!unixgetcurdir(&(path[3]), drive - 'A' + 1))
-			return(NULL);
+	else if (drive) {
+		file = gendospath(path, drive, _SC_);
+		drive = toupper2(drive);
+		if (drive == toupper2(*full)) strcpy(file, &(full[3]));
+		else if (!unixgetcurdir(file, drive - 'A' + 1)) return(NULL);
 	}
 #endif
 	else strcpy(path, full);
-
 	strcpy(strcatdelim(path), cp);
+
 	return(path);
 }
 
@@ -2183,6 +2264,7 @@ char *path, **dirp, *full;
 
 	if (!(cp = dostmpdir(drive))) return(-1);
 	*dirp = cp;
+
 	return(1);
 }
 #endif	/* !_NODOSDRIVE */
@@ -2226,7 +2308,7 @@ char *arc;
 
 	strcpy(path, arc);
 	if ((dest = strrdelim(path, 1))) *(++dest) = '\0';
-	else strcpy(path, ".");
+	else copycurpath(path);
 	if ((n = archdostmpdir(path, &tmpdest, full)) < 0) {
 		warning(ENOENT, arc);
 		return(0);
@@ -2243,7 +2325,7 @@ char *arc;
 		}
 	}
 #ifndef	_NODOSDRIVE
-	else if ((drive = tmpdosdupl("", &tmpdir, 0)) < 0) {
+	else if ((drive = tmpdosdupl(nullstr, &tmpdir, 0)) < 0) {
 		free(dest);
 		return(0);
 	}
@@ -2258,7 +2340,8 @@ char *arc;
 		duparchivefile = archivefile;
 		archivefile = NULL;
 		Xwaitmes();
-		ret = execusercomm(archivelist[i].p_comm, path, -1, 1, 0);
+		ret = execusercomm(archivelist[i].p_comm, path,
+			F_ARGSET | F_ISARCH | F_NOADDOPT);
 		if (ret > 0) {
 			if (ret < 127) {
 				hideclock = 1;
@@ -2272,7 +2355,7 @@ char *arc;
 	if (tmpdest) {
 		if (!ret) {
 			if (_chdir2(tmpdest) < 0) ret = 1;
-			else forcemovefile(dest);
+			else VOID_C forcemovefile(dest);
 		}
 		free(dest);
 		removetmp(tmpdest, NULL);
@@ -2286,10 +2369,10 @@ char *arc;
 }
 
 /*ARGSUSED*/
-int unpack(arc, dir, arg, tr, ignorelist)
+int unpack(arc, dir, arg, tr, flags)
 char *arc, *dir;
 char *arg;
-int tr, ignorelist;
+int tr, flags;
 {
 #ifndef	_NODOSDRIVE
 	winvartable *wvp;
@@ -2337,24 +2420,24 @@ int tr, ignorelist;
 #endif	/* !_NOTREE */
 		{
 			if (arg && *arg) dir = strdup2(arg);
-			else dir = inputstr(UNPAC_K, 1, -1, NULL, 1);
+			else dir = inputstr(UNPAC_K, 1, -1, NULL, HST_PATH);
 			dir = evalpath(dir, 0);
 		}
 		if (!dir) return(0);
-		if (!*dir) strcpy(path, ".");
+		if (!*dir) copycurpath(path);
 		else {
 			cp = strcpy2(path, dir);
 #ifdef	_USEDOSPATH
-			if (_dospath(dir) && !dir[2]) strcpy(cp, ".");
+			if (_dospath(dir) && !dir[2]) copycurpath(cp);
 #endif
 		}
 		free(dir);
 		dir = NULL;
 
 		if (!undertmp(path)) break;
-		locate(0, L_CMDLINE);
-		putterm(l_clear);
-		cprintf2("[%.*s]", n_column - 2, path);
+		Xlocate(0, L_CMDLINE);
+		Xputterm(L_CLEAR);
+		Xcprintf2("[%.*s]", n_column - 2, path);
 		if (yesno(UNPTM_K)) break;
 	}
 
@@ -2408,8 +2491,7 @@ int tr, ignorelist;
 #endif
 	else {
 		Xwaitmes();
-		ret = execusercomm(archivelist[i].u_comm, path,
-			-1, 1, ignorelist);
+		ret = execusercomm(archivelist[i].u_comm, path, flags);
 		if (ret > 0) {
 			if (ret < 127) {
 				hideclock = 1;
@@ -2421,7 +2503,7 @@ int tr, ignorelist;
 
 #ifndef	_NODOSDRIVE
 	if (tmpdest) {
-		if (!ret) forcemovefile(dest);
+		if (!ret) VOID_C forcemovefile(dest);
 		free(dest);
 		removetmp(tmpdest, NULL);
 	}
@@ -2429,6 +2511,7 @@ int tr, ignorelist;
 	else
 #endif	/* !_NODOSDRIVE */
 	if (_chdir2(fullpath) < 0) lostcwd(fullpath);
+
 	return((ret) ? 0 : 1);
 }
 
@@ -2454,7 +2537,8 @@ int single;
 		if (ismark(&(filelist[i]))) filelist[i].tmpflags |= F_WSMRK;
 	dupmark = mark;
 	if (single) mark = 0;
-	ret = unpack(archivefile, path, NULL, 0, 0);
+	ret = unpack(archivefile, path, NULL,
+		0, F_ARGSET | F_ISARCH | F_NOADDOPT);
 	mark = dupmark;
 	for (i = 0; i < maxfile; i++) {
 		if (wasmark(&(filelist[i]))) filelist[i].tmpflags |= F_ISMRK;
@@ -2464,7 +2548,7 @@ int single;
 	subdir = archivedir;
 	while (*subdir == _SC_) subdir++;
 
-	if (ret < 0) putterm(t_bell);
+	if (ret < 0) Xputterm(T_BELL);
 	else if (!ret) /*EMPTY*/;
 	else if (_chdir2(path) < 0) {
 		hideclock = 1;
@@ -2491,8 +2575,8 @@ int single;
 		}
 		if (i >= maxfile) return(tmpdir);
 	}
-
 	removetmp(tmpdir, NULL);
+
 	return(NULL);
 }
 
@@ -2515,7 +2599,7 @@ char *dev;
 	}
 
 	st.flags = 0;
-	if (!(tmp = evalcommand("tar cf %C %TA", dev, &st, 0))) return(0);
+	if (!(tmp = evalcommand("tar cf %C %TA", dev, &st))) return(0);
 
 #ifdef	_NOEXTRAMACRO
 	if (filelist) for (;;) {
@@ -2523,7 +2607,7 @@ char *dev;
 		free(tmp);
 
 		if (!(st.flags & F_REMAIN)
-		|| !(tmp = evalcommand("tar rf %C %TA", dev, NULL, 0)))
+		|| !(tmp = evalcommand("tar rf %C %TA", dev, NULL)))
 			break;
 	}
 	else
@@ -2539,6 +2623,7 @@ char *dev;
 		mark = 0;
 		marksize = (off_t)0;
 	}
+
 	return(1);
 }
 
@@ -2597,20 +2682,20 @@ int maxf, n;
 	else tmpdir = NULL;	/* fake for -Wuninitialized */
 #endif
 
-	locate(0, n_line - 1);
-	tflush();
-	if (!(fp = popenmacro(list -> comm, file, 0))) {
+	Xlocate(0, n_line - 1);
+	Xtflush();
+	if (!(fp = popenmacro(list -> comm, file, F_IGNORELIST))) {
 		if (n >= 0) removetmp(tmpdir, file);
 		return(0);
 	}
 
-	locate(0, L_MESLINE);
-	putterm(l_clear);
-	putterm(t_standout);
-	kanjiputs(SEACH_K);
-	putterm(end_standout);
-	kanjiputs(file);
-	tflush();
+	Xlocate(0, L_MESLINE);
+	Xputterm(L_CLEAR);
+	Xputterm(T_STANDOUT);
+	Xkanjiputs(SEACH_K);
+	Xputterm(END_STANDOUT);
+	Xkanjiputs(file);
+	Xtflush();
 
 	for (i = 0; i < (int)(list -> topskip); i++) {
 		if (!(cp = fgets2(fp, 0))) break;
