@@ -1,11 +1,12 @@
 /*
  *	file.c
  *
- *	File Access Module
+ *	file accessing module
  */
 
 #include <fcntl.h>
 #include "fd.h"
+#include "termio.h"
 #include "func.h"
 #include "kanji.h"
 
@@ -32,9 +33,6 @@ extern int checkdrive __P_((int));
 #endif
 
 extern int mark;
-extern off_t marksize;
-extern off_t totalsize;
-extern off_t blocksize;
 extern char fullpath[];
 extern char *tmpfilename;
 extern int physical_path;
@@ -61,7 +59,7 @@ extern int dosdrive;
 # else
 # define	L_SET	0
 # endif
-#endif
+#endif	/* !L_SET */
 
 #ifdef	_NODOSDRIVE
 #define	nodoschdir	Xchdir
@@ -78,6 +76,7 @@ static int NEAR cpfile __P_((char *, char *, struct stat *, struct stat *));
 static VOID changemes __P_((VOID_A));
 static int NEAR genrand __P_((int));
 static int dounlink __P_((char *));
+static int dormdir __P_((char *));
 #ifndef	_NOWRITEFS
 static int NEAR isexist __P_((char *));
 static int NEAR realdirsiz __P_((char *, int, int, int, int));
@@ -110,6 +109,7 @@ char *path, *file;
 	path[0] = '.';
 	path[1] = _SC_;
 	strcpy(&(path[2]), file);
+
 	return(path);
 }
 #endif	/* _USEDOSEMU */
@@ -135,6 +135,7 @@ gid_t gid;
 	else if (isgroupmember(gid)) mode >>= 3;
 #endif	/* !NOUID */
 	if (dir && !(mode & F_ISEXE)) mode &= ~(F_ISRED | F_ISWRI);
+
 	return(mode & 007);
 }
 
@@ -150,14 +151,15 @@ namelist *namep;
 	cp = nodospath(path, namep -> name);
 	if (Xlstat(cp, &lst) < 0 || stat2(cp, &st) < 0) return(-1);
 	namep -> flags = 0;
-	if ((st.st_mode & S_IFMT) == S_IFDIR) namep -> flags |= F_ISDIR;
-	if ((lst.st_mode & S_IFMT) == S_IFLNK) namep -> flags |= F_ISLNK;
+	if (s_isdir(&st)) namep -> flags |= F_ISDIR;
+	if (s_islnk(&lst)) namep -> flags |= F_ISLNK;
 
 	if (isdisplnk(dispmode))
 		memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
 #if	!MSDOS
 	if ((lst.st_mode & S_IFMT) == S_IFCHR
-	|| (lst.st_mode & S_IFMT) == S_IFBLK) namep -> flags |= F_ISDEV;
+	|| (lst.st_mode & S_IFMT) == S_IFBLK)
+		namep -> flags |= F_ISDEV;
 #endif
 
 	namep -> st_mode = lst.st_mode;
@@ -181,12 +183,8 @@ namelist *namep;
 #else
 		logical_access((u_int)(st.st_mode), st.st_uid, st.st_gid);
 #endif
-
-	if (isfile(namep) && namep -> st_size) {
-		totalsize += getblock(namep -> st_size);
-		if (ismark(namep)) marksize += getblock(namep -> st_size);
-	}
 	namep -> tmpflags |= F_STAT;
+
 	return(0);
 }
 
@@ -209,13 +207,8 @@ CONST VOID_P vp2;
 
 	if (!isdir(namep1) && isdir(namep2)) return(1);
 	if (isdir(namep1) && !isdir(namep2)) return(-1);
-
-	if (namep1 -> name[0] == '.' && namep1 -> name[1] == '\0') return(-1);
-	if (namep2 -> name[0] == '.' && namep2 -> name[1] == '\0') return(1);
-	if (namep1 -> name[0] == '.' && namep1 -> name[1] == '.'
-	&& namep1 -> name[2] == '\0') return(-1);
-	if (namep2 -> name[0] == '.' && namep2 -> name[1] == '.'
-	&& namep2 -> name[2] == '\0') return(1);
+	if ((tmp = isdotdir(namep2 -> name) - isdotdir(namep1 -> name)))
+		return(tmp);
 
 	switch (sorton & 7) {
 		case 5:
@@ -295,10 +288,10 @@ CONST VOID_P vp2;
 		case 2:
 			if ((cp1 = strrchr(((treelist *)vp1) -> name, '.')))
 				cp1++;
-			else cp1 = "";
+			else cp1 = nullstr;
 			if ((cp2 = strrchr(((treelist *)vp2) -> name, '.')))
 				cp2++;
-			else cp2 = "";
+			else cp2 = nullstr;
 			tmp = strpathcmp2(cp1, cp2);
 			break;
 		case 3:
@@ -307,8 +300,8 @@ CONST VOID_P vp2;
 			tmp = 0;
 			break;
 	}
-
 	if (sorton > 7) tmp = -tmp;
+
 	return(tmp);
 }
 #endif
@@ -331,8 +324,7 @@ char *arcregstr;
 	}
 #ifndef	_NOARCHIVE
 	else if (arcregstr) while ((dp = Xreaddir(dirp))) {
-		if (stat2(dp -> d_name, &st) < 0
-		|| (st.st_mode & S_IFMT) == S_IFDIR) continue;
+		if (stat2(dp -> d_name, &st) < 0 || s_isdir(&st)) continue;
 
 		tmp.name = dp -> d_name;
 		tmp.flags = 0;
@@ -343,6 +335,7 @@ char *arcregstr;
 	}
 #endif
 	else dp = Xreaddir(dirp);
+
 	return(dp);
 }
 
@@ -362,11 +355,10 @@ char *buf;
 	}
 	if (!homedir) {
 		if (!(cp = gethomedir())) return(-1);
-		if (
 #ifndef	_NODOSDRIVE
-		dospath2(cp) ||
+		if (dospath2(cp)) return(-1);
 #endif
-		_chdir2(cp) < 0) return(-1);
+		if (_chdir2(cp) < 0) return(-1);
 		homedir = getwd2();
 		if (_chdir2(cwd) < 0) lostcwd(cwd);
 	}
@@ -380,11 +372,12 @@ char *buf;
 #endif
 	else cp = underpath(cwd, homedir, len);
 
-	if (buf) strcpy(buf, (cp) ? cp : "");
+	if (buf) strcpy(buf, (cp) ? cp : nullstr);
 #ifdef	DEBUG
 	free(homedir);
 	homedir = NULL;
 #endif
+
 	return((cp) ? 1 : 0);
 }
 
@@ -416,25 +409,68 @@ char *dir;
 			}
 			else if (stat2(dir, &st) < 0) return(-1);
 		}
-		if ((st.st_mode & S_IFMT) != S_IFDIR) {
+		if (!s_isdir(&st)) {
 			errno = ENOTDIR;
 			return(-1);
 		}
 	}
-
 	entryhist(1, realpath2(dir, tmp, 0), 1);
+
 	return(0);
 }
 
+#ifndef	NOFLOCK
+int lockfile(fd, mode)
+int fd, mode;
+{
+	static int lockmode[] = {
+# ifdef	USEFCNTLOCK
+		F_RDLCK, F_WRLCK, F_UNLCK,
+# else	/* !USEFCNTLOCK */
+#  ifdef	USELOCKF
+		F_LOCK, F_LOCK, F_ULOCK,
+#  else
+		LOCK_SH, LOCK_EX, LOCK_UN,
+#  endif
+# endif	/* !USEFCNTLOCK */
+	};
+# ifdef	USEFCNTLOCK
+	struct flock lock;
+# endif
+
+# ifndef	_NODOSDRIVE
+	if (fd >= DOSFDOFFSET) return(0);
+# endif
+
+# ifdef	USEFCNTLOCK
+	lock.l_type = lockmode[mode];
+	lock.l_start = lock.l_len = (off_t)0;
+	lock.l_whence = SEEK_SET;
+	return(fcntl(fd, F_SETLKW, &lock));
+# else	/* !USEFCNTLOCK */
+#  ifdef	USELOCKF
+	return(lockf(fd, lockmode[mode], (off_t)0));
+#  else
+	return(flock(fd, lockmode[mode]));
+#  endif
+# endif	/* !USEFCNTLOCK */
+}
+#endif	/* !NOFLOCK */
+
 int touchfile(path, stp)
 char *path;
 struct stat *stp;
 {
-	struct stat st, dummy;
+#ifndef	_NOEXTRAATTR
+	int i;
+#endif
+	struct stat st;
+	u_int mode;
 	int ret, duperrno;
 
+	if (!(stp -> st_nlink)) return(0);
 	if (Xlstat(path, &st) < 0) return(-1);
-	if ((st.st_mode & S_IFMT) == S_IFLNK) return(1);
+	if (s_islnk(&st)) return(1);
 
 	ret = 0;
 #ifdef	FAKEUNINIT
@@ -443,10 +479,19 @@ struct stat *stp;
 #if	MSDOS
 	if (!(st.st_mode & S_IWRITE)) Xchmod(path, (st.st_mode | S_IWRITE));
 #endif
-	if (stp -> st_atime != (time_t)-1 && stp -> st_mtime != (time_t)-1) {
+
+	if (stp -> st_nlink & (TCH_ATIME | TCH_MTIME)) {
 #ifdef	USEUTIME
 		struct utimbuf times;
+#else
+		struct timeval tvp[2];
+#endif
 
+		if (!(stp -> st_nlink & TCH_ATIME))
+			stp -> st_atime = st.st_atime;
+		if (!(stp -> st_nlink & TCH_MTIME))
+			stp -> st_mtime = st.st_mtime;
+#ifdef	USEUTIME
 		times.actime = stp -> st_atime;
 		times.modtime = stp -> st_mtime;
 		if (Xutime(path, &times) < 0) {
@@ -454,8 +499,6 @@ struct stat *stp;
 			ret = -1;
 		}
 #else
-		struct timeval tvp[2];
-
 		tvp[0].tv_sec = stp -> st_atime;
 		tvp[0].tv_usec = 0;
 		tvp[1].tv_sec = stp -> st_mtime;
@@ -466,26 +509,36 @@ struct stat *stp;
 		}
 #endif
 	}
+
 #ifdef	HAVEFLAGS
-	if (stp -> st_flags != (u_long)-1) {
-# ifndef	_NODOSDRIVE
-		if (dospath(path, NULL)) {
-			duperrno = EACCES;
-			ret = -1;
-		}
-		else
-# endif
-		if (chflags(path, stp -> st_flags) < 0) {
+	if (stp -> st_nlink & TCH_FLAGS) {
+		if (Xchflags(path, stp -> st_flags) < 0) {
 			duperrno = errno;
 			ret = -1;
 		}
 	}
 #endif
-	dummy.st_mode = (u_short)-1;
-	if (stp -> st_mode != dummy.st_mode) {
-		stp -> st_mode &= ~S_IFMT;
-		stp -> st_mode |= (st.st_mode & S_IFMT);
-		if (Xchmod(path, stp -> st_mode) < 0) {
+
+	if (stp -> st_nlink & TCH_MODE) {
+		mode = stp -> st_mode;
+		mode &= ~S_IFMT;
+		mode |= (st.st_mode & S_IFMT);
+#ifndef	_NOEXTRAATTR
+		if ((stp -> st_nlink & TCH_MODEEXE)
+		&& !s_isdir(&st) && !(st.st_mode & S_IEXEC_ALL)) {
+			for (i = 0; i < 3; i++) {
+				if (!(stp -> st_mode & (1 << (i + 12))))
+					continue;
+				mode &= ~(1 << (i * 3));
+				mode |= (st.st_mode & (1 << (i * 3)));
+			}
+		}
+		if (stp -> st_nlink & TCH_MASK) {
+			mode &= ~(stp -> st_size);
+			mode |= (st.st_mode & (stp -> st_size));
+		}
+#endif	/* !_NOEXTRAATTR */
+		if (Xchmod(path, mode) < 0) {
 			duperrno = errno;
 			ret = -1;
 		}
@@ -493,17 +546,22 @@ struct stat *stp;
 #if	MSDOS
 	else if (!(st.st_mode & S_IWRITE)) Xchmod(path, st.st_mode);
 #endif
+
 #ifndef	NOUID
-	if (stp -> st_gid != (gid_t)-1) {
-# ifndef	_NODOSDRIVE
-		if (dospath(path, NULL)) /*EMPTY*/;
-		else
-# endif
-		if (chown(path, stp -> st_uid, stp -> st_gid) < 0)
-			chown(path, (uid_t)-1, stp -> st_gid);
+	if (stp -> st_nlink & (TCH_UID | TCH_GID)) {
+		if (!(stp -> st_nlink & TCH_UID)) stp -> st_uid = (uid_t)-1;
+		if (!(stp -> st_nlink & TCH_GID)) stp -> st_gid = (gid_t)-1;
+		if (Xchown(path, stp -> st_uid, stp -> st_gid) >= 0) /*EMPTY*/;
+		else if (!(stp -> st_nlink & TCH_CHANGE))
+			Xchown(path, (uid_t)-1, stp -> st_gid);
+		else {
+			duperrno = errno;
+			ret = -1;
+		}
 	}
 #endif	/* !NOUID */
 	if (ret < 0) errno = duperrno;
+
 	return(ret);
 }
 
@@ -517,6 +575,7 @@ char *path;
 	dosdrive = 0;
 	n = Xchdir(path);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -530,6 +589,7 @@ char *path;
 	dosdrive = 0;
 	cp = Xgetwd(path);
 	dosdrive = dupdosdrive;
+
 	return(cp);
 }
 
@@ -543,6 +603,7 @@ struct stat *stp;
 	dosdrive = 0;
 	n = Xlstat(path, stp);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -556,6 +617,7 @@ int mode;
 	dosdrive = 0;
 	n = Xmkdir(path, mode);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 
@@ -568,6 +630,7 @@ char *path;
 	dosdrive = 0;
 	n = Xrmdir(path);
 	dosdrive = dupdosdrive;
+
 	return(n);
 }
 #endif	/* !_NODOSDRIVE */
@@ -585,35 +648,50 @@ char *path;
 		cp = NOCWD_K;
 	else if ((cp = gethomedir()) && !nodoschdir(cp) && nodosgetwd(path))
 		cp = GOHOM_K;
-	else if (!nodoschdir(_SS_) && nodosgetwd(path)) cp = GOROT_K;
-	else error(_SS_);
+	else if (!nodoschdir(rootpath) && nodosgetwd(path)) cp = GOROT_K;
+	else error(rootpath);
 
 	if (path != fullpath) strncpy2(fullpath, path, MAXPATHLEN - 1);
 	warning(0, cp);
 #ifndef	_NOUSEHASH
-	searchhash(NULL, "", "");
+	searchhash(NULL, nullstr, nullstr);
 #endif
 	errno = duperrno;
 }
 
-int safewrite(fd, buf, size)
-int fd;
-char *buf;
-int size;
+#ifndef	NODIRLOOP
+int issamebody(src, dest)
+char *src, *dest;
 {
-	int n;
+	struct stat st1, st2;
 
-	n = Xwrite(fd, buf, size);
-#if	MSDOS
-	if (n >= 0 && n < size) {
-		n = -1;
-		errno = ENOSPC;
-	}
-#else
-	if (n < 0 && errno == EINTR) n = size;
-#endif
-	return(n);
+	if (Xstat(src, &st1) < 0 || Xstat(dest, &st2) < 0) return(0);
+
+	return(st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino);
 }
+#endif	/* !NODIRLOOP */
+
+#ifndef	NOSYMLINK
+int cpsymlink(src, dest)
+char *src, *dest;
+{
+	struct stat st;
+	char path[MAXPATHLEN];
+	int len;
+
+	if ((len = Xreadlink(src, path, strsize(path))) < 0) return(-1);
+	if (Xlstat(dest, &st) >= 0) {
+# ifndef	NODIRLOOP
+		if (issamebody(src, dest)) return(0);
+# endif
+		if (Xunlink(dest) < 0) return(-1);
+	}
+
+	path[len] = '\0';
+
+	return(Xsymlink(path, dest));
+}
+#endif	/* !NOSYMLINK */
 
 /*ARGSUSED*/
 static int NEAR cpfile(src, dest, stp1, stp2)
@@ -624,55 +702,64 @@ struct stat *stp1, *stp2;
 	struct stat st;
 #endif
 	char buf[BUFSIZ];
-	int i, fd1, fd2, duperrno;
+	int i, fd1, fd2, flags, mode, duperrno;
 
 #if	MSDOS
 	if (!stp2 && Xlstat(dest, &st) >= 0) stp2 = &st;
 	if (stp2 && !(stp2 -> st_mode & S_IWRITE))
 		Xchmod(src, stp2 -> st_mode | S_IWRITE);
 #endif
-	if ((stp1 -> st_mode & S_IFMT) == S_IFLNK) {
-		if ((i = Xreadlink(src, buf, BUFSIZ - 1)) < 0) return(-1);
-		buf[i] = '\0';
-		return(Xsymlink(buf, dest));
-	}
-	if ((fd1 = Xopen(src, O_BINARY | O_RDONLY, stp1 -> st_mode)) < 0)
-		return(-1);
-#if	MSDOS
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC,
-		stp1 -> st_mode | S_IWRITE);
-#else
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC, stp1 -> st_mode);
+#ifndef	NOSYMLINK
+	if (s_islnk(stp1)) return(cpsymlink(src, dest));
 #endif
-	if (fd2 < 0) {
-		Xclose(fd1);
+
+	flags = (O_BINARY | O_RDONLY);
+	mode = stp1 -> st_mode;
+	if ((fd1 = Xopen(src, flags, mode)) < 0) return(-1);
+
+	flags = (O_BINARY | O_WRONLY | O_CREAT | O_TRUNC);
+#if	MSDOS
+	mode |= S_IWRITE;
+#endif
+#ifndef	NODIRLOOP
+	if (issamebody(src, dest)) {
+		flags |= O_EXCL;
+		if (Xunlink(dest) < 0) {
+			VOID_C Xclose(fd1);
+			return(-1);
+		}
+	}
+#endif
+	if ((fd2 = Xopen(dest, flags, mode)) < 0) {
+		VOID_C Xclose(fd1);
 		return(-1);
 	}
 
+#ifdef	FAKEUNINIT
+	duperrno = errno;	/* fake for -Wuninitialized */
+#endif
 	for (;;) {
-		while ((i = Xread(fd1, buf, BUFSIZ)) < 0 && errno == EINTR);
-		duperrno = errno;
-		if (i < BUFSIZ) break;
-		if ((i = safewrite(fd2, buf, BUFSIZ)) < 0) {
+		if ((i = sureread(fd1, buf, BUFSIZ)) <= 0) {
 			duperrno = errno;
 			break;
 		}
+		if (surewrite(fd2, buf, i) < 0) {
+			i = -1;
+			duperrno = errno;
+			break;
+		}
+		if (i < BUFSIZ) break;
 	}
-	if (i > 0 && (i = safewrite(fd2, buf, i)) < 0) duperrno = errno;
 
-	Xclose(fd2);
-	Xclose(fd1);
+	VOID_C Xclose(fd2);
+	VOID_C Xclose(fd1);
 
 	if (i < 0) {
 		Xunlink(dest);
 		errno = duperrno;
 		return(-1);
 	}
-#ifdef	HAVEFLAGS
-	stp1 -> st_flags = (u_long)-1;
-#endif
+	stp1 -> st_nlink = (TCH_ATIME | TCH_MTIME);
 #ifdef	_USEDOSCOPY
 	if (touchfile(dest, stp1) < 0) return(-1);
 #else
@@ -680,6 +767,7 @@ struct stat *stp1, *stp2;
 	if (inheritcopy && touchfile(dest, stp1) < 0) return(-1);
 # endif
 #endif
+
 	return(0);
 }
 
@@ -723,13 +811,12 @@ char *src, *dest;
 struct stat *stp1, *stp2;
 {
 	if (Xrename(src, dest) >= 0) return(0);
-	if (errno != EXDEV || (stp1 -> st_mode & S_IFMT) == S_IFDIR)
-		return(-1);
+	if (errno != EXDEV || s_isdir(stp1)) return(-1);
 	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
 		return(-1);
-#ifdef	HAVEFLAGS
-	stp1 -> st_flags = (u_long)-1;
-#endif
+
+	stp1 -> st_nlink = (TCH_MODE | TCH_UID | TCH_GID
+		| TCH_ATIME | TCH_MTIME);
 	return (touchfile(dest, stp1));
 }
 
@@ -771,8 +858,8 @@ int len;
 	int i, j, c;
 
 	if (!buf) {
-		for (i = 0; i < sizeof(seq) / sizeof(char); i++) {
-			j = genrand(sizeof(seq) / sizeof(char));
+		for (i = 0; i < arraysize(seq); i++) {
+			j = genrand(arraysize(seq));
 			c = seq[i];
 			seq[i] = seq[j];
 			seq[j] = c;
@@ -780,11 +867,12 @@ int len;
 	}
 	else {
 		for (i = 0; i < len; i++) {
-			j = genrand(sizeof(seq) / sizeof(char));
+			j = genrand(arraysize(seq));
 			buf[i] = seq[j];
 		}
 		buf[i] = '\0';
 	}
+
 	return(buf);
 }
 
@@ -823,9 +911,9 @@ char *dir;
 		}
 	}
 	if (!tmpfilename) {
-		n = sizeof(TMPPREFIX) - 1;
+		n = strsize(TMPPREFIX);
 		strncpy(cp, TMPPREFIX, n);
-		len = sizeof(path) - 1 - (cp - path);
+		len = strsize(path) - (cp - path);
 		if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 		len -= n;
 		genrandname(NULL, 0);
@@ -844,7 +932,7 @@ char *dir;
 	}
 
 	strncpy((cp = strcatdelim(path)), dir, n);
-	len = sizeof(path) - 1 - (cp - path);
+	len = strsize(path) - (cp - path);
 	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 	len -= n;
 	genrandname(NULL, 0);
@@ -864,6 +952,7 @@ char *dir;
 		nodosrmdir(path);
 		errno = n;
 	}
+
 	return(-1);
 }
 
@@ -884,6 +973,7 @@ char *dir;
 	}
 	else if (errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
 		return(-1);
+
 	return(0);
 }
 
@@ -896,7 +986,7 @@ char *file;
 	path[0] = '\0';
 	if (mktmpdir(path) < 0) return(-1);
 	cp = strcatdelim(path);
-	len = sizeof(path) - 1 - (cp - path);
+	len = strsize(path) - (cp - path);
 	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
 	genrandname(NULL, 0);
 
@@ -913,6 +1003,7 @@ char *file;
 	duperrno = errno;
 	rmtmpdir(NULL);
 	errno = duperrno;
+
 	return(-1);
 }
 
@@ -921,20 +1012,25 @@ char *file;
 {
 	if ((Xunlink(file) < 0 && errno != ENOENT) || rmtmpdir(NULL) < 0)
 		return(-1);
+
 	return(0);
 }
 
 static int dounlink(path)
 char *path;
 {
-	return(Xunlink(path));
+	if (Xunlink(path) < 0) return(APL_ERROR);
+
+	return(APL_OK);
 }
 
 static int dormdir(path)
 char *path;
 {
-	if (isdotdir(path)) return(0);
-	return(Xrmdir(path));
+	if (isdotdir(path)) return(APL_OK);
+	if (Xrmdir(path) < 0) return(APL_ERROR);
+
+	return(APL_OK);
 }
 
 VOID removetmp(dir, file)
@@ -949,7 +1045,8 @@ char *dir, *file;
 		warning(-1, dir);
 		*dir = '\0';
 	}
-	else if (!file) applydir(NULL, dounlink, NULL, dormdir, 3, NULL);
+	else if (!file) VOID_C applydir(NULL, dounlink,
+		NULL, dormdir, ORD_NOPREDIR, NULL);
 	else if (Xunlink(nodospath(path, file)) < 0) warning(-1, file);
 
 #ifndef	_NODOSDRIVE
@@ -986,7 +1083,7 @@ char *dir, *file;
 	strcatdelim2(buf, dir, file);
 
 	if (rawchdir(buf) != 0) return(0);
-	rawchdir(_SS_);
+	rawchdir(rootpath);
 #if	MSDOS
 	spawnlpe(P_WAIT, "DELTREE.EXE", "DELTREE", "/Y", buf, NULL, environ);
 #else
@@ -996,6 +1093,7 @@ char *dir, *file;
 		_exit(1);
 	}
 #endif
+
 	return(0);
 }
 
@@ -1009,6 +1107,7 @@ int drive;
 	path[1] = drive;
 	path[2] = '\0';
 	if (mktmpdir(path) < 0) return(NULL);
+
 	return(strdup2(path));
 }
 
@@ -1063,6 +1162,7 @@ int single;
 		return(-1);
 	}
 	*dirp = tmpdir;
+
 	return(drive);
 }
 
@@ -1076,9 +1176,7 @@ char *file;
 	char tmp[MAXPATHLEN];
 # endif
 
-	path[0] = drive;
-	path[1] = ':';
-	strcpy(&(path[2]), file);
+	strcpy(gendospath(path, drive, '\0'), file);
 # if	!MSDOS
 	file = nodospath(tmp, file);
 # endif
@@ -1097,6 +1195,7 @@ char *file;
 	struct stat st;
 
 	if (Xlstat(file, &st) < 0 && errno == ENOENT) return(0);
+
 	return(1);
 }
 
@@ -1134,6 +1233,7 @@ int dos, boundary, dirsize, ofs;
 		len += ofs;
 		if (lfn) return((len / LFNENTSIZ + 1) * DOSDIRENT + DOSDIRENT);
 	}
+
 	return(DOSDIRENT);
 }
 
@@ -1149,6 +1249,7 @@ int size, dos, boundary, dirsize, ofs;
 		size -= DOSDIRENT;
 		return((size / DOSDIRENT) * LFNENTSIZ - 1);
 	}
+
 	return(DOSBODYLEN);
 }
 
@@ -1164,6 +1265,7 @@ char *from, *to;
 	from = nodospath(fpath, from);
 	to = nodospath(tpath, to);
 #endif
+
 	return(Xrename(from, to));
 }
 
@@ -1210,9 +1312,10 @@ char *tmpdir, *old;
 					else break;
 				}
 				else {
-					if ((fd = Xopen(fname,
-					O_CREAT | O_EXCL, 0600))) {
-						Xclose(fd);
+					fd = Xopen(fname,
+						O_CREAT | O_EXCL, 0600);
+					if (fd >= 0) {
+						VOID_C Xclose(fd);
 						return(fname);
 					}
 					if (errno != EEXIST) break;
@@ -1225,6 +1328,7 @@ char *tmpdir, *old;
 		}
 	}
 	free(fname);
+
 	return(NULL);
 }
 
@@ -1235,6 +1339,7 @@ char *dir;
 	struct stat st;
 
 	if (Xlstat(dir, &st) < 0) return(getblocksize(dir));
+
 	return((off_t)st.st_size);
 }
 
@@ -1244,7 +1349,7 @@ off_t bsiz;
 {
 	struct stat st;
 	u_char ch, *tmp;
-	int i, c, n, fd;
+	int i, n, fd;
 
 	if (Xlstat(dir, &st) < 0
 	|| (fd = Xopen(dir, O_BINARY | O_RDONLY, 0666)) < 0)
@@ -1257,16 +1362,14 @@ off_t bsiz;
 			free(tmp);
 			return(NULL);
 		}
-		while ((c = Xread(fd, (char *)&ch, sizeof(ch))) < 0
-		&& errno == EINTR);
-		if (c < 0) {
+		if (sureread(fd, &ch, sizeof(ch)) < 0) {
 			free(tmp);
 			return(NULL);
 		}
 		tmp[i] = ch + 1;
 	}
 	tmp[i] = 0;
-	Xclose(fd);
+	VOID_C Xclose(fd);
 
 	return(tmp);
 }
@@ -1407,7 +1510,7 @@ int fs;
 	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	waitmes();
 
-	if (!(dirp = Xopendir("."))) {
+	if (!(dirp = Xopendir(curpath))) {
 		freevar(fnamelist);
 		noconv--;
 		lostcwd(path);
@@ -1457,8 +1560,8 @@ int fs;
 
 #if	!MSDOS
 	if (fs != FSID_EFS) entnum = NULL;	/* except IRIX File System */
-	else if (!(entnum = getentnum(".", persec))) {
-		warning(-1, ".");
+	else if (!(entnum = getentnum(curpath, persec))) {
+		warning(-1, curpath);
 		restorefile(tmpdir, path, fnamp);
 		freevar(fnamelist);
 		noconv--;
@@ -1466,8 +1569,8 @@ int fs;
 	}
 	totalent = headbyte
 		+ realdirsiz(tmpdir, dos, boundary, dirsize, namofs)
-		+ realdirsiz(".", dos, boundary, dirsize, namofs)
-		+ realdirsiz("..", dos, boundary, dirsize, namofs);
+		+ realdirsiz(curpath, dos, boundary, dirsize, namofs)
+		+ realdirsiz(parentpath, dos, boundary, dirsize, namofs);
 	block = tmpno = 0;
 	ptr = 3;
 	totalptr = 0;
@@ -1532,7 +1635,8 @@ int fs;
 		free(tmpdir);
 		tmpdir = cp;
 	}
-	fnamp = strcatdelim2(path, tmpdir, fnamelist[top]) - path;
+	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
+	snprintf2(&(path[fnamp]), sizeof(path) - fnamp, fnamelist[top]);
 	if (saferename(path, fnamelist[top]) < 0) warning(-1, path);
 	restorefile(tmpdir, path, fnamp);
 
