@@ -4,7 +4,6 @@
  *	by T.Shirai <shirai@unixusers.net>
  */
 
-#include <signal.h>
 #include "fd.h"
 #include "func.h"
 #include "kanji.h"
@@ -15,7 +14,7 @@
 # ifndef	DJGPP
 # include <dos.h>
 #  ifdef	__TURBOC__
-extern unsigned _stklen = 0x5800;
+extern u_int _stklen = 0x5800;
 #  define	harderr_t	void
 #  else
 #  define	harderr_t	int
@@ -24,9 +23,17 @@ extern unsigned _stklen = 0x5800;
 #endif	/* !MSDOS */
 
 #ifdef	_NOORIGSHELL
+#include "termio.h"
+#include "wait.h"
 #define	Xexit2		exit2
+#define	isorgpid()	(1)
 #else
 #include "system.h"
+#define	isorgpid()	(mypid == orgpid)
+#endif
+
+#ifndef	DATADIR
+#define	DATADIR		progpath
 #endif
 
 #if	MSDOS
@@ -53,11 +60,10 @@ extern int maxlaunch;
 extern archivetable archivelist[];
 extern int maxarchive;
 #endif
+extern int wheader;
 extern char fullpath[];
-extern char *histfile;
+extern int savehist;
 extern char *helpindex[];
-extern int tradlayout;
-extern int sizeinfo;
 extern int subwindow;
 extern int win_x;
 extern int win_y;
@@ -65,35 +71,17 @@ extern int win_y;
 extern int custno;
 #endif
 extern char *deftmpdir;
-#ifndef	_NODOSDRIVE
+#ifdef	_USEUNICODE
 extern char *unitblpath;
 #endif
+#ifndef	_NOPTY
+extern int ptymode;
+extern int ptyinternal;
+extern int parentfd;
+#endif
 
 #define	CLOCKUPDATE	10	/* sec */
 
-#ifndef	SIG_DFL
-#define	SIG_DFL		((sigcst_t)0)
-#endif
-#ifndef	SIG_IGN
-#define	SIG_IGN		((sigcst_t)1)
-#endif
-
-#if	!MSDOS && !defined (SIGWINCH)
-# if	defined (SIGWINDOW)
-# define	SIGWINCH	SIGWINDOW
-# else
-#  if	defined (NSIG)
-#  define	SIGWINCH	NSIG
-#  else
-#  define	SIGWINCH	28
-#  endif
-# endif
-#endif	/* !MSDOS && !SIGWINCH */
-
-#if	!defined (SIGIOT) && defined (SIGABRT)
-#define	SIGIOT	SIGABRT
-#endif
-
 #if	MSDOS && !defined (PROTECTED_MODE)
 static harderr_t far criticalerror __P_((u_int, u_int, u_short far *));
 #endif
@@ -152,11 +140,21 @@ static int xcpuerror __P_((VOID_A));
 #ifdef	SIGXFSZ
 static int xsizerror __P_((VOID_A));
 #endif
+static int NEAR minrow __P_((int));
+#ifndef	_NOEXTRAWIN
+static int NEAR biaswsize __P_((int, int));
+#endif
 #ifdef	SIGWINCH
 static int wintr __P_((VOID_A));
 #endif
+static VOID NEAR printtime __P_((int));
 #ifdef	SIGALRM
-static int printtime __P_((VOID_A));
+static int trapalrm __P_((VOID_A));
+#endif
+static char *NEAR getversion __P_((int *));
+#ifndef	_NOLOGGING
+static VOID NEAR startlog __P_((char **));
+static VOID NEAR endlog __P_((int));
 #endif
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline __P_((char *, char *, int, char *));
@@ -166,7 +164,7 @@ static int NEAR evaloption __P_((char *[]));
 static char *NEAR searchenv __P_((char *, char *[]));
 static VOID NEAR setexecname __P_((char *));
 static VOID NEAR setexecpath __P_((char *, char *[]));
-static VOID NEAR prepareexitfd __P_((VOID_A));
+int main __P_((int, char *[], char *[]));
 
 char *origpath = NULL;
 char *progpath = NULL;
@@ -176,11 +174,13 @@ char *tmpfilename = NULL;
 int adjtty = 0;
 #endif
 int showsecond = 0;
+#if	FD >= 2
+int thruargs = 0;
+#endif
 int hideclock = 0;
 #ifdef	SIGALRM
 int noalrm = 1;
 #endif
-u_short today[3] = {0, 0, 0};
 int fd_restricted = 0;
 #ifndef	_NOCUSTOMIZE
 char **orighelpindex = NULL;
@@ -205,8 +205,7 @@ int fdmode = 0;
 
 static int timersec = 0;
 #ifdef	SIGWINCH
-static int winchok = 0;
-static int winched = 0;
+static int nowinch = 0;
 #endif
 
 
@@ -228,13 +227,16 @@ char *s;
 
 	duperrno = errno;
 	sigvecset(0);
-	forcecleandir(deftmpdir, tmpfilename);
+	if (isorgpid()) {
+		forcecleandir(deftmpdir, tmpfilename);
 #ifndef	_NODOSDRIVE
-	dosallclose();
+		dosallclose();
 #endif
+		Xstdiomode();
+		endterm();
+	}
+
 	if (!s) s = progname;
-	stdiomode();
-	endterm();
 	if (dumbterm <= 2) fputc('\007', stderr);
 	errno = duperrno;
 	if (errno) perror2(s);
@@ -242,26 +244,28 @@ char *s;
 		fputs(s, stderr);
 		fputnl(stderr);
 	}
-	inittty(1);
-	keyflush();
-	prepareexitfd();
+
+	if (isorgpid()) {
+		inittty(1);
+		keyflush();
+		prepareexitfd(2);
 #ifndef	_NOORIGSHELL
 # ifndef	NOJOB
-	killjob();
+		killjob();
 # endif
-	prepareexit(0);
+		prepareexit(0);
+#endif	/* !_NOORIGSHELL */
+#ifndef	_NOPTY
+		killallpty();
 #endif
 #ifdef	DEBUG
 # if	!MSDOS
-	freeterment();
+		freeterment();
 # endif
-	if (ttyout && ttyout != stderr) {
-		if (fileno(ttyout) == ttyio) ttyio = -1;
-		fclose(ttyout);
+		closetty(&ttyio, &ttyout);
+		muntrace();
+#endif	/* DEBUG */
 	}
-	if (ttyio >= 0) close(ttyio);
-	muntrace();
-#endif
 
 	exit(2);
 }
@@ -270,12 +274,36 @@ static VOID NEAR signalexit(sig)
 int sig;
 {
 	signal2(sig, SIG_IGN);
-	forcecleandir(deftmpdir, tmpfilename);
+
+	if (isorgpid()) {
+		forcecleandir(deftmpdir, tmpfilename);
 #ifndef	_NODOSDRIVE
-	dosallclose();
+		dosallclose();
 #endif
-	endterm();
-	inittty(1);
+#ifndef	_NOLOGGING
+		endlog(sig + 128);
+#endif
+		endterm();
+		inittty(1);
+		keyflush();
+#ifndef	_NOORIGSHELL
+# if	defined (SIGHUP) && !defined (NOJOB)
+		if (sig == SIGHUP) killjob();
+# endif
+		prepareexit(0);
+#endif	/* !_NOORIGSHELL */
+#ifndef	_NOPTY
+		killallpty();
+#endif
+#ifdef	DEBUG
+# if	!MSDOS
+		freeterment();
+# endif
+		closetty(&ttyio, &ttyout);
+		muntrace();
+#endif	/* DEBUG */
+	}
+
 	signal2(sig, SIG_DFL);
 	VOID_C kill(getpid(), sig);
 }
@@ -283,50 +311,79 @@ int sig;
 #ifdef	SIGALRM
 static int ignore_alrm(VOID_A)
 {
+	int duperrno;
+
+	duperrno = errno;
 	signal2(SIGALRM, (sigcst_t)ignore_alrm);
+	errno = duperrno;
+
 	return(0);
 }
-#endif
+#endif	/* SIGALRM */
 
 #ifdef	SIGWINCH
 static int ignore_winch(VOID_A)
 {
+	int duperrno;
+
+	duperrno = errno;
 	signal2(SIGWINCH, (sigcst_t)ignore_winch);
+	errno = duperrno;
+
 	return(0);
 }
-#endif
+#endif	/* SIGWINCH */
 
 #ifdef	SIGINT
 static int ignore_int(VOID_A)
 {
+	int duperrno;
+
+	duperrno = errno;
 	signal2(SIGINT, (sigcst_t)ignore_int);
+	errno = duperrno;
+
 	return(0);
 }
-#endif
+#endif	/* SIGINT */
 
 #ifdef	SIGQUIT
 static int ignore_quit(VOID_A)
 {
+	int duperrno;
+
+	duperrno = errno;
 	signal2(SIGQUIT, (sigcst_t)ignore_quit);
+	errno = duperrno;
+
 	return(0);
 }
-#endif
+#endif	/* SIGQUIT */
 
 #ifdef	SIGCONT
 static int ignore_cont(VOID_A)
 {
+	int duperrno;
+
+	duperrno = errno;
 	signal2(SIGCONT, (sigcst_t)ignore_cont);
 # if	!MSDOS
 	suspended = 1;
-# endif
+#  ifndef	_NOORIGSHELL
+	sigconted = 1;
+#  endif
+# endif	/* !MSDOS */
+	errno = duperrno;
+
 	return(0);
 }
-#endif
+#endif	/* SIGCONT */
 
 #ifdef	SIGHUP
 static int hanguperror(VOID_A)
 {
 	signalexit(SIGHUP);
+
 	return(0);
 }
 #endif
@@ -335,6 +392,7 @@ static int hanguperror(VOID_A)
 static int illerror(VOID_A)
 {
 	signalexit(SIGILL);
+
 	return(0);
 }
 #endif
@@ -343,6 +401,7 @@ static int illerror(VOID_A)
 static int traperror(VOID_A)
 {
 	signalexit(SIGTRAP);
+
 	return(0);
 }
 #endif
@@ -351,6 +410,7 @@ static int traperror(VOID_A)
 static int ioerror(VOID_A)
 {
 	signalexit(SIGIOT);
+
 	return(0);
 }
 #endif
@@ -359,6 +419,7 @@ static int ioerror(VOID_A)
 static int emuerror(VOID_A)
 {
 	signalexit(SIGEMT);
+
 	return(0);
 }
 #endif
@@ -367,6 +428,7 @@ static int emuerror(VOID_A)
 static int floaterror(VOID_A)
 {
 	signalexit(SIGFPE);
+
 	return(0);
 }
 #endif
@@ -375,6 +437,7 @@ static int floaterror(VOID_A)
 static int buserror(VOID_A)
 {
 	signalexit(SIGBUS);
+
 	return(0);
 }
 #endif
@@ -383,6 +446,7 @@ static int buserror(VOID_A)
 static int segerror(VOID_A)
 {
 	signalexit(SIGSEGV);
+
 	return(0);
 }
 #endif
@@ -391,6 +455,7 @@ static int segerror(VOID_A)
 static int syserror(VOID_A)
 {
 	signalexit(SIGSYS);
+
 	return(0);
 }
 #endif
@@ -399,6 +464,7 @@ static int syserror(VOID_A)
 static int pipeerror(VOID_A)
 {
 	signalexit(SIGPIPE);
+
 	return(0);
 }
 #endif
@@ -407,6 +473,7 @@ static int pipeerror(VOID_A)
 static int terminate(VOID_A)
 {
 	signalexit(SIGTERM);
+
 	return(0);
 }
 #endif
@@ -415,6 +482,7 @@ static int terminate(VOID_A)
 static int xcpuerror(VOID_A)
 {
 	signalexit(SIGXCPU);
+
 	return(0);
 }
 #endif
@@ -423,143 +491,240 @@ static int xcpuerror(VOID_A)
 static int xsizerror(VOID_A)
 {
 	signalexit(SIGXFSZ);
+
 	return(0);
 }
 #endif
 
+VOID setlinecol(VOID_A)
+{
+	char buf[MAXLONGWIDTH + 1];
+
+	if (getconstvar("LINES")) {
+		snprintf2(buf, sizeof(buf), "%d", n_line);
+		setenv2("LINES", buf, 1);
+	}
+	if (getconstvar("COLUMNS")) {
+		snprintf2(buf, sizeof(buf), "%d", n_column);
+		setenv2("COLUMNS", buf, 1);
+	}
+}
+
+static int NEAR minrow(n)
+int n;
+{
+	if (n == win) {
+#ifndef	_NOTREE
+		if (treepath) return(WFILEMINTREE);
+#endif
+#ifndef	_NOCUSTOMIZE
+		if (custno >= 0) return(WFILEMINCUSTOM);
+#endif
+	}
+
+	return(WFILEMIN);
+}
+
+#ifndef	_NOEXTRAWIN
+static int NEAR biaswsize(n, max)
+int n, max;
+{
+	int i, row, min;
+
+	for (i = row = 0; i < n; i++) row += winvar[i].v_fileperrow + 1;
+	min = minrow(n);
+	if (max - row < min) {
+		winvar[n].v_fileperrow = min;
+		return(max - min);
+	}
+	winvar[n].v_fileperrow = max - row;
+
+	return(0);
+}
+#endif	/* !_NOEXTRAWIN */
+
 VOID checkscreen(xmax, ymax)
 int xmax, ymax;
 {
-#ifdef	SIGWINCH
-	int dupwinchok;
-#endif
-	char *cp;
-	int i, row;
+	char *cp, *tty;
+	int i, row, wastty, dupn_line, dupdumbterm;
 
 #ifdef	SIGWINCH
-	dupwinchok = winchok;
-	winchok = 0;
+	nowinch++;
 #endif
+	dupn_line = -1;
+#ifndef	_NOPTY
+	if (parentfd >= 0) row = WFILEMIN;
+	else
+#endif
+	{
+		row = wheader + WFOOTER;
+#ifdef	_NOEXTRAWIN
+		row += (minrow(win) + 1) * windows;
+#else
+		for (i = 0; i < windows; i++) row += minrow(i) + 1;
+#endif
+		row--;
+	}
+
+	savetermio(ttyio, &tty, NULL);
+	wastty = isttyiomode;
+	Xttyiomode(1);
+	dupdumbterm = dumbterm;
+	if (xmax < 0 || ymax < 0) {
+		dupn_line = n_line;
+		xmax = WCOLUMNMIN;
+		ymax = 0;
+		dumbterm = 1;
+	}
 	for (i = 0;; i++) {
 		if (!(cp = getwsize(xmax, ymax))) {
-#ifndef	_NOTREE
-			if (treepath) row = WFILEMINTREE;
-			else
-#endif
-#ifndef	_NOCUSTOMIZE
-			if (custno >= 0) row = WFILEMINCUSTOM;
-			else
-#endif
-			row = WFILEMIN;
-
-			if (FILEPERROW >= row) break;
+			if (n_line >= row) break;
 			cp = NOROW_K;
 		}
 		if (!i) {
-			putterm(t_clear);
-			locate(0, 0);
+			Xputterms(T_CLEAR);
+			Xlocate(0, 0);
 			keyflush();
 		}
 
-		if (i & 1) cputs2(SCRSZ_K);
+		if (i & 1) Xcputs2(SCRSZ_K);
 		else {
-			putterm(t_standout);
-			cputs2(cp);
-			putterm(end_standout);
+			Xputterm(T_STANDOUT);
+			Xcputs2(cp);
+			Xputterm(END_STANDOUT);
 		}
-		putterm(t_bell);
-		cputnl();
-		keyflush();
+		Xputterm(T_BELL);
+		Xcputnl();
+		Xtflush();
 		if (kbhit2(1000000L) && getkey2(0) == K_ESC) {
 			errno = 0;
 			error(INTR_K);
 		}
 	}
+	dumbterm = dupdumbterm;
+	loadtermio(ttyio, tty, NULL);
+	isttyiomode = wastty;
+	if (tty) free(tty);
+
+	setlinecol();
+	if (n_line != dupn_line) {
+#ifdef	_NOEXTRAWIN
+		calcwin();
+#else
+		row = fileperrow(1);
+		for (i = windows - 1; i >= 0; i--) {
+			if (!(row = biaswsize(i, row))) break;
+			row--;
+		}
+#endif
+	}
 #ifdef	SIGWINCH
-	winchok = dupwinchok;
+	nowinch--;
 #endif
 }
 
 #ifdef	SIGWINCH
+VOID pollscreen(forced)
+int forced;
+{
+	static int winched = 0;
+	int x, y;
+
+	if (forced < 0) nowinch = winched = 0;
+	else if (nowinch) {
+		if (forced) winched++;
+	}
+	else if (winched || forced) {
+		winched = 0;
+		x = n_column;
+		y = n_line;
+		checkscreen(-1, -1);
+		if (isorgpid()) {
+			if (x != n_column || y != n_line) rewritefile(1);
+			if (subwindow) ungetch2(K_CTRL('L'));
+		}
+	}
+}
+
 static int wintr(VOID_A)
 {
-	int duperrno, dupusegetcursor;
+	int duperrno;
 
 	duperrno = errno;
 	signal2(SIGWINCH, SIG_IGN);
-	if (!winchok) winched++;
-	else {
-		dupusegetcursor = usegetcursor;
-		usegetcursor = 0;
-		checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
-		usegetcursor = dupusegetcursor;
-		rewritefile(1);
-		if (subwindow) ungetch2(K_CTRL('L'));
-	}
+	pollscreen(1);
 	signal2(SIGWINCH, (sigcst_t)wintr);
 	errno = duperrno;
+
 	return(0);
 }
-#endif
+#endif	/* SIGWINCH */
 
-#ifdef	SIGALRM
-static int printtime(VOID_A)
+static VOID NEAR printtime(hide)
+int hide;
 {
 	static time_t now;
 	struct tm *tm;
-	int x, duperrno;
+	int x;
 
-	duperrno = errno;
-	signal2(SIGALRM, SIG_IGN);
 	if (timersec) now++;
 	else {
 		now = time2();
 		timersec = CLOCKUPDATE;
 	}
-	if (showsecond || timersec == CLOCKUPDATE) {
-# ifdef	DEBUG
-		_mtrace_file = "localtime(start)";
-		tm = localtime(&now);
-		if (_mtrace_file) _mtrace_file = NULL;
-		else {
-			_mtrace_file = "localtime(end)";
-			malloc(0);	/* dummy malloc */
-		}
-# else
-		tm = localtime(&now);
-# endif
-		today[0] = tm -> tm_year;
-		today[1] = tm -> tm_mon;
-		today[2] = tm -> tm_mday;
-		if (!hideclock) {
-			x = n_column - 15 - ((showsecond) ? 3 : 0);
-			if (!isleftshift()) x--;
-			locate(x, L_TITLE);
-			putterm(t_standout);
-			cprintf2("%02d-%02d-%02d %02d:%02d",
-				tm -> tm_year % 100,
-				tm -> tm_mon + 1, tm -> tm_mday,
-				tm -> tm_hour, tm -> tm_min);
-			if (showsecond) cprintf2(":%02d", tm -> tm_sec);
-			putterm(end_standout);
-			locate(win_x, win_y);
-			tflush();
-		}
+	if (timersec-- < CLOCKUPDATE && !showsecond) return;
+#ifndef	_NOPTY
+	if (checkallpty() < 0) rewritefile(0);
+	if (isptymode() && !ptyinternal && hide <= 1) hide = 0;
+#endif
+	if (hide) return;
+
+#ifdef	DEBUG
+	_mtrace_file = "localtime(start)";
+	tm = localtime(&now);
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "localtime(end)";
+		malloc(0);	/* dummy malloc */
 	}
-	timersec--;
+#else
+	tm = localtime(&now);
+#endif
+	x = n_column - 15 - ((showsecond) ? 3 : 0);
+	if (!isleftshift()) x--;
+	Xlocate(x, L_TITLE);
+	Xputterm(T_STANDOUT);
+	Xcprintf2("%02d-%02d-%02d %02d:%02d",
+		tm -> tm_year % 100,
+		tm -> tm_mon + 1, tm -> tm_mday,
+		tm -> tm_hour, tm -> tm_min);
+	if (showsecond) Xcprintf2(":%02d", tm -> tm_sec);
+	Xputterm(END_STANDOUT);
+	Xlocate(win_x, win_y);
+	Xtflush();
+}
+
+#ifdef	SIGALRM
+static int trapalrm(VOID_A)
+{
+	int duperrno;
+
+	duperrno = errno;
+	signal2(SIGALRM, SIG_IGN);
+# if	!defined (_NOORIGSHELL) && !defined (_NOPTY)
+	if (isshptymode()) {
+		if (mypid == shellpid) printtime(hideclock);
+	}
+	else
+# endif
+	if (isorgpid()) printtime(hideclock);
 # ifdef	SIGWINCH
-	if (!winchok) winchok++;
-	else if (winched) {
-		signal2(SIGWINCH, SIG_IGN);
-		winched = 0;
-		checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
-		rewritefile(1);
-		if (subwindow) ungetch2(K_CTRL('L'));
-		signal2(SIGWINCH, (sigcst_t)wintr);
-	}
+	pollscreen(0);
 # endif
-	signal2(SIGALRM, (sigcst_t)printtime);
+	signal2(SIGALRM, (sigcst_t)trapalrm);
 	errno = duperrno;
+
 	return(0);
 }
 #endif	/* SIGALRM */
@@ -574,14 +739,13 @@ int set;
 	if (set == old) /*EMPTY*/;
 	else if (set) {
 #ifdef	SIGALRM
-		signal2(SIGALRM, (sigcst_t)printtime);
+		signal2(SIGALRM, (sigcst_t)trapalrm);
 		noalrm = 0;
 #endif
 #ifdef	SIGTSTP
 		signal2(SIGTSTP, SIG_IGN);
 #endif
 #ifdef	SIGWINCH
-		winchok = winched = 0;
 		signal2(SIGWINCH, (sigcst_t)wintr);
 #endif
 #ifndef	_NODOSDRIVE
@@ -611,50 +775,60 @@ int set;
 	return(old);
 }
 
-VOID title(VOID_A)
+static char *NEAR getversion(lenp)
+int *lenp;
 {
 	char *cp, *eol;
-	int i, len;
 
-	locate(0, L_TITLE);
-	putterm(t_standout);
-	len = 0;
-	if (!isleftshift()) {
-		putch2(' ');
-		len++;
-	}
-	cputs2(" FD");
-	len += 3;
-	if (!ishardomit()) {
-		cputs2("(File & Directory tool)");
-		len += 23;
-	}
-	cputs2(" Ver.");
-	len += 5;
 	cp = strchr(version, ' ');
 	while (*(++cp) == ' ');
-	if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
-	i = eol - cp;
-	cprintf2("%-*.*s", i, i, cp);
+	if (lenp) {
+		if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
+		*lenp = eol - cp;
+	}
+
+	return(cp);
+}
+
+VOID title(VOID_A)
+{
+	char *cp;
+	int i, len;
+
+	Xlocate(0, L_TITLE);
+	Xputterm(T_STANDOUT);
+	len = 0;
+	if (!isleftshift()) {
+		Xputch2(' ');
+		len++;
+	}
+	Xcputs2(" FD");
+	len += 3;
+	if (!ishardomit()) {
+		Xcputs2("(File & Directory tool)");
+		len += 23;
+	}
+	Xcputs2(" Ver.");
+	len += 5;
+	cp = getversion(&i);
+	cputstr(i, cp);
 	if (distributor) {
-		putch2('#');
+		Xputch2('#');
 		i++;
 	}
-	cp = (iswellomit()) ? "" : " (c)1995-2004 T.Shirai  ";
-	cputs2(cp);
+	cp = (iswellomit()) ? nullstr : " (c)1995-2006 T.Shirai  ";
+	Xcputs2(cp);
 	i = n_column - len - strlen2(cp) - i;
-	while (i-- > 0) putch2(' ');
-	putterm(end_standout);
+	while (i-- > 0) Xputch2(' ');
+	Xputterm(END_STANDOUT);
 	timersec = 0;
-#ifdef	SIGALRM
-	printtime();
-#endif
+	printtime(0);
 }
 
 #ifndef	_NOCUSTOMIZE
 VOID saveorigenviron(VOID_A)
 {
-	orighelpindex = copystrarray(NULL, helpindex, NULL, 10);
+	orighelpindex = copystrarray(NULL, helpindex, NULL, MAXHELPINDEX);
 	origbindlist = copybind(NULL, bindlist);
 # ifndef	_NOKEYMAP
 	origkeymaplist = copykeyseq(NULL);
@@ -671,6 +845,61 @@ VOID saveorigenviron(VOID_A)
 }
 #endif	/* !_NOCUSTOMIZE */
 
+#ifndef	_NOLOGGING
+static VOID NEAR startlog(argv)
+char **argv;
+{
+# ifndef	NOUID
+	uid_t uid;
+	int lvl;
+# endif
+	char *cp, *tmp, buf[MAXLOGLEN + 1];
+	int i, len;
+
+	cp = buf;
+	for (i = 1; argv && argv[i]; i++) {
+		if (cp > buf) *(cp++) = ' ';
+		tmp = killmeta(argv[i]);
+		len = snprintf2(cp, (int)sizeof(buf) - (cp - buf), "%s", tmp);
+		free(tmp);
+		if (len < 0) break;
+		cp += len;
+	}
+	*cp = '\0';
+
+	cp = getversion(&len);
+# ifdef	NOUID
+	logmessage(_LOG_DEBUG_, "%s (%-*.*s) starts; PWD=%k; ARGS=%k",
+		progname, len, len, cp, origpath, buf);
+# else
+	lvl = ((uid = getuid())) ? _LOG_DEBUG_ : _LOG_WARNING_;
+	logmessage(lvl, "%s (%-*.*s) starts; UID=%d; PWD=%k; ARGS=%k",
+		progname, len, len, cp, uid, origpath, buf);
+# endif
+}
+
+static VOID NEAR endlog(status)
+int status;
+{
+# ifndef	NOUID
+	uid_t uid;
+	int lvl;
+# endif
+	char cwd[MAXPATHLEN];
+
+	if (!Xgetwd(cwd)) strcpy(cwd, "?");
+# ifdef	NOUID
+	logmessage(_LOG_DEBUG_, "%s ends; PWD=%k; STATUS=%d",
+		progname, cwd, status);
+# else
+	lvl = ((uid = getuid())) ? _LOG_DEBUG_ : _LOG_WARNING_;
+	logmessage(lvl, "%s ends; UID=%d; PWD=%k; STATUS=%d",
+		progname, uid, cwd, status);
+# endif
+	logclose();
+}
+#endif	/* !_NOLOGGING */
+
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline(command, file, n, line)
 char *command, *file;
@@ -681,16 +910,17 @@ char *line;
 	int i;
 
 	if (!*(cp = skipspace(command))) i = 0;
-	else i = execpseudoshell(cp, 0, 1);
+	else i = execpseudoshell(cp, F_IGNORELIST | F_NOCOMLINE);
 	if (i) {
-		putterm(l_clear);
-		cprintf2("%s, line %d: %s", file, n, ILFNC_K);
-		cputnl();
-		putterm(l_clear);
-		cprintf2("\t%s", line);
-		cputnl();
+		Xputterm(L_CLEAR);
+		Xcprintf2("%s, line %d: %s", file, n, ILFNC_K);
+		Xcputnl();
+		Xputterm(L_CLEAR);
+		Xcprintf2("\t%s", line);
+		Xcputnl();
 	}
 	free(command);
+
 	return((i) ? -1 : 0);
 }
 #endif	/* _NOORIGSHELL */
@@ -716,7 +946,7 @@ int exist;
 		cp = malloc2(strlen(file) + strlen(tmp) + 1 + 1);
 		strcpy(strcpy2(strcpy2(cp, file), "."), tmp);
 		tmp = evalpath(cp, 0);
-		if (stat2(tmp, &st) < 0 || (st.st_mode & S_IFMT) != S_IFREG) {
+		if (stat2(tmp, &st) < 0 || !s_isreg(&st)) {
 			free(tmp);
 			tmp = NULL;
 		}
@@ -730,8 +960,8 @@ int exist;
 			free(tmp);
 			return(0);
 		}
-		cprintf2("%s: Not found", tmp);
-		cputnl();
+		Xcprintf2("%s: Not found", tmp);
+		Xcputnl();
 		free(tmp);
 		return(-1);
 	}
@@ -777,9 +1007,9 @@ int exist;
 	fclose(fp);
 	free(tmp);
 #else	/* !_NOORIGSHELL */
-	n = stdiomode();
+	if ((n = isttyiomode)) Xstdiomode();
 	er = execruncom(file, 1);
-	ttyiomode(n);
+	if (n) Xttyiomode(n - 1);
 #endif	/* !_NOORIGSHELL */
 
 	return(er ? -1 : 0);
@@ -800,7 +1030,8 @@ char *argv[], *envp[];
 	for (i = 1; i < argc; i++) {
 		if (argv[i][0] == '+') /*EMPTY*/;
 		else if (argv[i][0] != '-' || !argv[i][1]
-		|| (argv[i][1] == '-' && !argv[i][2])) break;
+		|| (argv[i][1] == '-' && !argv[i][2]))
+			break;
 		else if (!isidentchar(argv[i][1])) /*EMPTY*/;
 		else {
 			for (cp = &(argv[i][2]); *cp; cp++)
@@ -819,6 +1050,7 @@ char *argv[], *envp[];
 	if (initshell(optc, optv) < 0) Xexit2(RET_FAIL);
 #endif	/* !_NOORIGSHELL */
 	free(optv);
+
 	return(argc);
 }
 
@@ -829,14 +1061,17 @@ char *argv[];
 	int i;
 
 	for (i = 1; argv[i]; i++) {
-		if (argv[i][0] != '-') return(i);
-		if (!argv[i][1] || (argv[i][1] == '-' && !argv[i][2]))
-			return(i + 1);
+		if (argv[i][0] != '-') break;
+		if (!argv[i][1] || (argv[i][1] == '-' && !argv[i][2])) {
+			i++;
+			break;
+		}
 		tmp = strdup2(&(argv[i][1]));
 		if ((cp = strchr(tmp, '='))) *(cp++) = '\0';
 		setenv2(tmp, cp, 0);
 		free(tmp);
 	}
+
 	return(i);
 }
 
@@ -849,6 +1084,7 @@ char *s, *envp[];
 	for (i = 0; envp[i]; i++)
 		if (!strnenvcmp(envp[i], s, len) && envp[i][len] == '=')
 			return(&(envp[i][++len]));
+
 	return(NULL);
 }
 
@@ -902,18 +1138,56 @@ char *argv, *envp[];
 	}
 }
 
-static VOID NEAR prepareexitfd(VOID_A)
+/*ARGSUSED*/
+VOID initfd(argv)
+char **argv;
+{
+#ifndef	_NOORIGSHELL
+	if (!interactive) /*EMPTY*/;
+	else
+#endif
+	{
+#if	!MSDOS
+		if (adjtty) {
+			Xstdiomode();
+			inittty(0);
+			Xttyiomode(0);
+		}
+#endif	/* !MSDOS */
+		loadhistory(0);
+		entryhist(1, origpath, 1);
+	}
+#ifndef	_NOLOGGING
+	startlog(argv);
+#endif
+}
+
+/*ARGSUSED*/
+VOID prepareexitfd(status)
+int status;
 {
 	char cwd[MAXPATHLEN];
 
+#ifndef	_NOLOGGING
+	endlog(status);
+#endif
+#ifndef	_NOORIGSHELL
+	if (!interactive) /*EMPTY*/;
+	else
+#endif
+	savehistory(0);
+
 	cwd[0] = '\0';
 	if (origpath && _chdir2(origpath) < 0) {
 		perror2(origpath);
 		if (!Xgetwd(cwd)) *cwd = '\0';
-		rawchdir(_SS_);
+		rawchdir(rootpath);
 	}
 	free(origpath);
 	free(progname);
+#ifdef	_USEUNICODE
+	free(unitblpath);
+#endif
 #ifndef	_NODOSDRIVE
 	dosallclose();
 #endif
@@ -927,7 +1201,7 @@ static VOID NEAR prepareexitfd(VOID_A)
 # endif
 # ifndef	_NOCUSTOMIZE
 	if (orighelpindex) {
-		freestrarray(orighelpindex, 10);
+		freestrarray(orighelpindex, MAXHELPINDEX);
 		free(orighelpindex);
 	}
 	free(origbindlist);
@@ -965,7 +1239,7 @@ static VOID NEAR prepareexitfd(VOID_A)
 # ifndef	_NOROCKRIDGE
 	detranspath(NULL, NULL);
 # endif
-# if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
+# ifdef	_USEUNICODE
 	discardunitable();
 # endif
 #endif	/* DEBUG */
@@ -1069,7 +1343,8 @@ char *argv[], *envp[];
 	}
 #endif	/* !_NOARCHIVE */
 
-	for (i = 0; i < 10; i++) helpindex[i] = strdup2(helpindex[i]);
+	for (i = 0; i < MAXHELPINDEX; i++)
+		helpindex[i] = strdup2(helpindex[i]);
 #ifdef	_USEDOSEMU
 	for (i = 0; fdtype[i].name; i++) {
 		fdtype[i].name = strdup2(fdtype[i].name);
@@ -1086,27 +1361,30 @@ char *argv[], *envp[];
 	environ[i] = NULL;
 #endif
 
+	initenv();
 	setexecpath(argv[0], envp);
-#ifndef	_NODOSDRIVE
-# ifdef	DATADIR
-	unitblpath = DATADIR;
-# else
-	unitblpath = progpath;
-# endif
+#ifdef	_USEUNICODE
+	unitblpath = strdup2(DATADIR);
 #endif
 
 #ifdef	_NOORIGSHELL
 	inittty(0);
 	getterment(NULL);
 	argc = initoption(argc, argv, envp);
+	adjustpath();
 #else	/* !_NOORIGSHELL */
 	cp = getshellname(progname, NULL, NULL);
 	if (!strpathcmp(cp, FDSHELL) || !strpathcmp(cp, "su")) {
 		i = main_shell(argc, argv, envp);
-		prepareexitfd();
+# ifndef	_NOPTY
+		killallpty();
+# endif
 		Xexit2(i);
 	}
-	fdmode = interactive = 1;
+#ifndef	_NOORIGSHELL
+	interactive =
+#endif
+	fdmode = 1;
 	setshellvar(envp);
 	prepareterm();
 	argc = initoption(argc, argv, envp);
@@ -1116,65 +1394,57 @@ char *argv[], *envp[];
 	}
 #endif	/* !_NOORIGSHELL */
 
-	ttyiomode(0);
+	Xttyiomode(0);
 	initterm();
 	if ((cp = getwsize(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN))) {
 		errno = 0;
 		error(cp);
 	}
+	setlinecol();
 #ifndef	_NOCUSTOMIZE
 	saveorigenviron();
 #endif
 
-	locate(0, n_line - 1);
+	Xlocate(0, n_line - 1);
 	inruncom = 1;
-#if	MSDOS && !defined (BSPATHDELIM)
-	/* DEFRUNCOM is gave from command line, not to convert previously */
-/*NOTDEFINED*/
-	cp = strdup2(DEFRUNCOM);
-	i = loadruncom(adjustpname(cp), 0);
-	free(cp);
-#else
-/*NOTDEFINED*/
-	i = loadruncom(DEFRUNCOM, 0);
-#endif
-	i += loadruncom(RUNCOMFILE, 0);
+	i = loadruncom(DEFRC, 0);
+	i += loadruncom(FD_RCFILE, 0);
 	inruncom = 0;
+#ifdef	SIGWINCH
+	nowinch = 1;
+#endif
 	sigvecset(1);
 	if (i < 0) {
-		hideclock = 1;
+		hideclock = 2;
 		warning(0, HITKY_K);
 	}
-	i = evaloption(argv);
-#ifdef	_NOORIGSHELL
-	adjustpath();
-#endif
-	evalenv();
-#if	!MSDOS
-	if (adjtty) {
-		stdiomode();
-		inittty(0);
-		ttyiomode(0);
-	}
-#endif	/* !MSDOS */
 
-	loadhistory(0, histfile);
-	entryhist(1, origpath, 1);
-	putterms(t_clear);
+	i = evaloption(argv);
+	checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
+	initfd(argv);
+	Xputterms(T_CLEAR);
 
 #ifdef	SIGWINCH
-	winchok = 1;
+	nowinch = 0;
+#endif
+#if	FD >= 2
+	main_fd(&(argv[i]), (thruargs) ? 1 : 0);
+#else
+	main_fd(&(argv[i]), 0);
 #endif
-	main_fd(argv[i]);
 	sigvecset(0);
-	prepareexitfd();
+#ifndef	_NOLOGGING
+	endlog(0);
+#endif
 
-	stdiomode();
-#ifndef	_NOORIGSHELL
-# ifndef	NOJOB
+	Xstdiomode();
+#if	!defined (_NOORIGSHELL) && !defined (NOJOB)
 	killjob();
-# endif
+#endif
+#ifndef	_NOPTY
+	killallpty();
 #endif
 	Xexit2(0);
+
 	return(0);
 }
