@@ -1,19 +1,16 @@
 /*
  *	pathname.c
  *
- *	Path Name Management Module
+ *	path name management module
  */
 
 #ifdef	FD
 #include "fd.h"
 #else	/* !FD */
+#include "machine.h"
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
-#include "machine.h"
-#include "printf.h"
-#include "kctype.h"
-#include "pathname.h"
 
 # ifndef	NOUNISTDH
 # include <unistd.h>
@@ -22,6 +19,10 @@
 # ifndef	NOSTDLIBH
 # include <stdlib.h>
 # endif
+
+#include "printf.h"
+#include "kctype.h"
+#include "pathname.h"
 #endif	/* !FD */
 
 #if	MSDOS
@@ -96,6 +97,7 @@ extern int Xclosedir __P_((DIR *));
 extern struct dirent *Xreaddir __P_((DIR *));
 extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
+extern int stat2 __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
 # if	MSDOS && !defined (_NOUSELFN)
 extern char *shortname __P_((char *, char *));
@@ -121,12 +123,23 @@ struct dirent *Xreaddir __P_((DIR *));
 static u_int NEAR getdosmode __P_((u_int));
 static time_t NEAR getdostime __P_((u_int, u_int));
 int Xstat __P_((char *, struct stat *));
+#  ifndef	NOSYMLINK
+#  define	Xlstat		Xstat
+#  endif
 # else	/* !MSDOS */
 # define	Xopendir	opendir
 # define	Xclosedir	closedir
 # define	Xreaddir	readdir
-# define	Xstat		stat
+# define	Xstat(p, s)	((stat(p, s)) ? -1 : 0)
+#  ifndef	NOSYMLINK
+#  define	Xlstat(p, s)	((lstat(p, s)) ? -1 : 0)
+#  endif
 # endif	/* !MSDOS */
+# ifdef	NOSYMLINK
+# define	stat2		Xstat
+# else
+int stat2 __P_((char *, struct stat *));
+# endif
 # ifdef	DJGPP
 char *Xgetwd __P_((char *));
 # else	/* !DJGPP */
@@ -136,20 +149,20 @@ char *Xgetwd __P_((char *));
 #  define	Xgetwd(p)	(char *)getcwd(p, MAXPATHLEN)
 #  endif
 # endif	/* !DJGPP */
-#define	Xaccess		access
+#define	Xaccess(p, m)	((access(p, m)) ? -1 : 0)
 #endif	/* !FD */
 
 static char *NEAR getenvvar __P_((char *, int));
 static int NEAR setvar __P_((char *, char *, int));
-static int NEAR ismeta __P_((char *s, int, int, int));
+static int NEAR isescape __P_((char *s, int, int, int, int));
 #ifdef	_NOORIGGLOB
 static char *NEAR cnvregexp __P_((char *, int));
 #else
 static int NEAR _regexp_exec __P_((char **, char *));
 #endif
 static void NEAR addstrbuf __P_((strbuf_t *, char *, int));
-static void NEAR duplwild(wild_t *, wild_t *);
-static void NEAR freewild(wild_t *);
+static void NEAR duplwild __P_((wild_t *, wild_t *));
+static void NEAR freewild __P_((wild_t *));
 static int NEAR _evalwild __P_((int, char ***, wild_t *));
 #ifndef	_NOUSEHASH
 static int NEAR calchash __P_((char *));
@@ -164,14 +177,11 @@ static int NEAR isexecute __P_((char *, int, int));
 static int NEAR extaccess __P_((char *, char *, int, int));
 #endif
 #if	!defined (FDSH) && !defined (_NOCOMPLETE)
-# ifndef	NOUID
-static int NEAR completeuser __P_((char *, int, int, char ***));
-# endif
 static int NEAR completefile __P_((char *, int, int, char ***,
 	char *, int, int));
 static int NEAR completeexe __P_((char *, int, int, char ***));
 #endif	/* !FDSH && !_NOCOMPLETE */
-static int NEAR addmeta __P_((char *, char *, int, int));
+static int NEAR addmeta __P_((char *, char *, int));
 static int NEAR skipvar __P_((char **, int *, int *, int));
 #ifdef	MINIMUMSHELL
 static int NEAR skipvarvalue __P_((char *, int *, char *, int, int));
@@ -194,6 +204,10 @@ static char *NEAR replacebackquote __P_((char *, int *, char *, int));
 			: (type *)realloc2(ptr, b_size(n, type)))
 #define	getconstvar(s)	(getenvvar(s, sizeof(s) - 1))
 
+char nullstr[] = "";
+char rootpath[] = _SS_;
+char curpath[] = ".";
+char parentpath[] = "..";
 char **argvar = NULL;
 #ifndef	_NOUSEHASH
 hashlist **hashtable = NULL;
@@ -234,6 +248,7 @@ char *s;
 	fputs(s, stderr);
 	fputs(": memory allocation error", stderr);
 	fputnl(stderr);
+
 	exit(2);
 }
 
@@ -248,6 +263,7 @@ ALLOC_T size;
 		tmp = NULL;	/* fake for -Wuninitialized */
 #endif
 	}
+
 	return(tmp);
 }
 
@@ -265,6 +281,7 @@ ALLOC_T size;
 		tmp = NULL;	/* fake for -Wuninitialized */
 #endif
 	}
+
 	return(tmp);
 }
 
@@ -276,7 +293,9 @@ ALLOC_T n, *sizep;
 		*sizep = BUFUNIT;
 		return(malloc2(*sizep));
 	}
-	while (n + 1 >= *sizep) *sizep *= 2;
+	n++;
+	while (n >= *sizep) *sizep *= 2;
+
 	return(realloc2(ptr, *sizep));
 }
 
@@ -290,6 +309,7 @@ char *s;
 	n = strlen(s);
 	if (!(tmp = (char *)malloc((ALLOC_T)n + 1))) error("malloc()");
 	memcpy(tmp, s, n + 1);
+
 	return(tmp);
 }
 
@@ -305,6 +325,7 @@ int n;
 	if (!(tmp = (char *)malloc((ALLOC_T)i + 1))) error("malloc()");
 	memcpy(tmp, s, i);
 	tmp[i] = '\0';
+
 	return(tmp);
 }
 
@@ -321,6 +342,7 @@ int c;
 		else if (isekana(s, i)) i++;
 #endif
 	}
+
 	return(NULL);
 }
 
@@ -331,6 +353,7 @@ char *s1, *s2;
 
 	for (i = 0; s2[i]; i++) s1[i] = s2[i];
 	s1[i] = '\0';
+
 	return(&(s1[i]));
 }
 
@@ -342,6 +365,7 @@ int n;
 
 	for (i = 0; i < n && s2[i]; i++) s1[i] = s2[i];
 	s1[i] = '\0';
+
 	return(&(s1[i]));
 }
 
@@ -383,6 +407,7 @@ char *dir;
 			return(NULL);
 		}
 	}
+
 	return(dirp);
 }
 
@@ -392,6 +417,7 @@ DIR *dirp;
 	free(dirp -> dd_buf);
 	free(dirp -> dd_path);
 	free(dirp);
+
 	return(0);
 }
 
@@ -465,10 +491,32 @@ struct stat *stp;
 	stp -> st_dev = stp -> st_ino = 0;
 	stp -> st_nlink = 1;
 	stp -> st_uid = stp -> st_gid = -1;
+
 	return(0);
 }
 # endif	/* MSDOS */
 
+# ifndef	NOSYMLINK
+int stat2(path, stp)
+char *path;
+struct stat *stp;
+{
+	int duperrno;
+
+	if (Xstat(path, stp) < 0) {
+		duperrno = errno;
+		if (Xlstat(path, stp) < 0
+		|| (stp -> st_mode & S_IFMT) != S_IFLNK) {
+			errno = duperrno;
+			return(-1);
+		}
+		stp -> st_mode &= ~S_IFMT;
+	}
+
+	return(0);
+}
+# endif	/* !NOSYMLINK */
+
 # ifdef	DJGPP
 char *Xgetwd(path)
 char *path;
@@ -478,12 +526,25 @@ char *path;
 
 	if (!(cp = (char *)getcwd(path, MAXPATHLEN))) return(NULL);
 	for (i = 0; cp[i]; i++) if (cp[i] == '/') cp[i] = _SC_;
+
 	return(cp);
 }
 # endif	/* !DJGPP */
 #endif	/* !FD */
 
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+char *gendospath(path, drive, c)
+char *path;
+int drive, c;
+{
+	*(path++) = drive;
+	*(path++) = ':';
+	if (c) *(path++) = c;
+	*path = '\0';
+
+	return(path);
+}
+
 char *strdelim(s, d)
 char *s;
 int d;
@@ -497,6 +558,7 @@ int d;
 		if (iskanji1(s, i)) i++;
 # endif
 	}
+
 	return(NULL);
 }
 
@@ -515,6 +577,7 @@ int d;
 		if (iskanji1(s, i)) i++;
 # endif
 	}
+
 	return(cp);
 }
 #endif	/* MSDOS || (FD && !_NODOSDRIVE) */
@@ -531,9 +594,11 @@ char *s, *eol;
 		if (s[i] == _SC_) cp = &(s[i]);
 		if (iskanji1(s, i)) i++;
 	}
+
 	return(cp);
 #else
 	for (eol--; eol >= s; eol--) if (*eol == _SC_) return(eol);
+
 	return(NULL);
 #endif
 }
@@ -553,6 +618,7 @@ int ptr;
 
 	for (i = 0; s[i] && i < ptr; i++) if (iskanji1(s, i)) i++;
 	if (!s[i] || i > ptr) return(1);
+
 	return(!iskanji1(s, i));
 #else
 	return(1);
@@ -590,6 +656,7 @@ char *s;
 		*cp = _SC_;
 	}
 	*(++cp) = '\0';
+
 	return(cp);
 }
 
@@ -643,6 +710,7 @@ char *buf, *s1, *s2;
 		for (i = 0; s2[i] && i < len; i++) *(cp++) = s2[i];
 	}
 	*cp = '\0';
+
 	return(cp);
 }
 
@@ -662,6 +730,7 @@ char *s1, *s2;
 		s1++;
 		s2++;
 	}
+
 	return(0);
 }
 
@@ -683,6 +752,7 @@ int n;
 		s1++;
 		s2++;
 	}
+
 	return(0);
 }
 
@@ -691,7 +761,8 @@ int strpathcmp2(s1, s2)
 char *s1, *s2;
 {
 	if (pathignorecase) return(strcasecmp2(s1, s2));
-	else return(strcmp(s1, s2));
+
+	return(strcmp(s1, s2));
 }
 
 int strnpathcmp2(s1, s2, n)
@@ -699,7 +770,8 @@ char *s1, *s2;
 int n;
 {
 	if (pathignorecase) return(strncasecmp2(s1, s2, n));
-	else return(strncmp(s1, s2, n));
+
+	return(strncmp(s1, s2, n));
 }
 #endif	/* !PATHNOCASE */
 
@@ -733,6 +805,7 @@ int len;
 	cp = strndup2(ident, len);
 	env = getenv2(cp);
 	free(cp);
+
 	return(env);
 #endif
 }
@@ -756,6 +829,7 @@ int len;
 	ret = (putvarfunc) ? (*putvarfunc)(cp, len) : putenv(cp);
 #endif
 	if (ret < 0) free(cp);
+
 	return(ret);
 }
 
@@ -774,7 +848,11 @@ int c;
 int isdotdir(s)
 char *s;
 {
-	if (s[0] == '.' && (!s[1] || (s[1] == '.' && !s[2]))) return(1);
+	if (s[0] != '.') /*EMPTY*/;
+	else if (!s[1]) return(2);
+	else if (s[1] != '.') /*EMPTY*/;
+	else if (!s[2]) return(1);
+
 	return(0);
 }
 
@@ -785,15 +863,37 @@ char *s;
 	if (_dospath(s)) s += 2;
 #endif
 	if (*s != _SC_) return(NULL);
+
 	return(s);
 }
 
+int isrootpath(s)
+char *s;
+{
+	return((s[0] == _SC_ && !s[1]));
+}
+
+VOID copyrootpath(s)
+char *s;
+{
+	*(s++) = _SC_;
+	*s = '\0';
+}
+
+VOID copycurpath(s)
+char *s;
+{
+	*(s++) = '.';
+	*s = '\0';
+}
+
 char *getbasename(s)
 char *s;
 {
 	char *cp;
 
 	if ((cp = strrdelim(s, 1))) return(cp + 1);
+
 	return(s);
 }
 
@@ -820,9 +920,9 @@ int *loginp, *restrictedp;
 	return(s);
 }
 
-static int NEAR ismeta(s, ptr, quote, len)
+static int NEAR isescape(s, ptr, quote, len, flags)
 char *s;
-int ptr, quote, len;
+int ptr, quote, len, flags;
 {
 #ifdef	FAKEMETA
 	return(0);
@@ -830,7 +930,7 @@ int ptr, quote, len;
 	if (s[ptr] != PMETA || quote == '\'') return(0);
 
 	if (len >= 0) {
-		if (ptr + 1 >= len) return(0);
+		if (!(flags & EA_EOLMETA) && ptr + 1 >= len) return(0);
 # ifndef	BASHSTYLE
 	/* bash does not treat "\" as \ */
 		if (quote == '"' && s[ptr + 1] == quote && ptr + 2 >= len)
@@ -838,7 +938,7 @@ int ptr, quote, len;
 # endif
 	}
 	else {
-		if (!s[ptr + 1]) return(0);
+		if (!(flags & EA_EOLMETA) && !s[ptr + 1]) return(0);
 # ifndef	BASHSTYLE
 	/* bash does not treat "\" as \ */
 		if (quote == '"' && s[ptr + 1] == quote && !s[ptr + 2])
@@ -884,7 +984,7 @@ int len;
 			re[j++] = s[i];
 			continue;
 		}
-		else if (ismeta(s, i, '\0', len)) {
+		else if (isescape(s, i, '\0', len, 0)) {
 			re[j++] = s[i++];
 			re[j++] = s[i];
 		}
@@ -913,6 +1013,7 @@ int len;
 	}
 	re[j++] = '$';
 	re[j++] = '\0';
+
 	return(re);
 }
 #endif	/* _NOORIGGLOB */
@@ -935,6 +1036,7 @@ int len;
 	s = cnvregexp(s, len);
 	re_comp(s);
 	free(s);
+
 	return((reg_t *)1);
 }
 
@@ -945,6 +1047,7 @@ char *s;
 int fname;
 {
 	if (fname && skipdotfile && *s == '.') return(0);
+
 	return(re_exec(s) > 0);
 }
 
@@ -976,6 +1079,7 @@ int len;
 		re = NULL;
 	}
 	free(s);
+
 	return(re);
 }
 
@@ -985,6 +1089,7 @@ char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
+
 	return(!regexec(re, s, 0, NULL, 0));
 }
 
@@ -1012,6 +1117,7 @@ int len;
 	s = cnvregexp(s, len);
 	re = regcmp(s, 0);
 	free(s);
+
 	return(re);
 }
 
@@ -1021,6 +1127,7 @@ char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
+
 	return(regex(re, s) ? 1 : 0);
 }
 
@@ -1053,7 +1160,7 @@ int len;
 		metachar = 0;
 		pc = parsechar(&(s[i]), len - i, '\0', 0, &quote, NULL);
 		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) continue;
-		else if (pc == PC_META) {
+		else if (pc == PC_ESCAPE) {
 			metachar = 1;
 			i++;
 		}
@@ -1196,6 +1303,7 @@ int len;
 		re = b_realloc(re, n, reg_t);
 		re[n] = NULL;
 	}
+
 	return((reg_t *)realloc2(re, (n + 1) * sizeof(reg_t)));
 }
 
@@ -1265,6 +1373,7 @@ char *s;
 	}
 	while (re[n1] == wildsymbol2) n1++;
 	if (re[n1] || s[n2]) return(0);
+
 	return(1);
 }
 
@@ -1274,6 +1383,7 @@ char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
+
 	return(_regexp_exec(re, s));
 }
 
@@ -1307,7 +1417,7 @@ int len;
 static void NEAR duplwild(dst, src)
 wild_t *dst, *src;
 {
-	memcpy(dst, src, sizeof(wild_t));
+	memcpy((char *)dst, (char *)src, sizeof(wild_t));
 	dst -> fixed.s = (char *)malloc2(src -> fixed.size);
 	memcpy(dst -> fixed.s, src -> fixed.s, src -> fixed.len + 1);
 	dst -> path.s = (char *)malloc2(src -> path.size);
@@ -1317,7 +1427,8 @@ wild_t *dst, *src;
 	if (src -> ino) {
 		dst -> ino = (devino_t *)malloc2(src -> nino
 			* sizeof(devino_t));
-		memcpy(dst -> ino, src -> ino, src -> nino * sizeof(devino_t));
+		memcpy((char *)(dst -> ino), (char *)(src -> ino),
+			src -> nino * sizeof(devino_t));
 	}
 #endif	/* !NODIRLOOP */
 }
@@ -1359,7 +1470,7 @@ wild_t *wp;
 	plen = wp -> path.len;
 	quote = wp -> quote;
 
-	if (wp -> fixed.len) addstrbuf(&(wp -> path), _SS_, 1);
+	if (wp -> fixed.len) addstrbuf(&(wp -> path), rootpath, 1);
 
 	for (i = w = 0; wp -> s[i] && wp -> s[i] != _SC_; i++) {
 		pc = parsechar(&(wp -> s[i]), -1,
@@ -1374,7 +1485,7 @@ wild_t *wp;
 			addstrbuf(&(wp -> path), &(wp -> s[i]), 1);
 			i++;
 		}
-		else if (pc == PC_META) {
+		else if (pc == PC_ESCAPE) {
 			if (wp -> flags & EA_KEEPMETA)
 				addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
 			if (wp -> s[i + 1] == _SC_) continue;
@@ -1401,12 +1512,12 @@ wild_t *wp;
 	if (!(wp -> s[i])) isdir = 0;
 	else {
 		isdir = 1;
-		addstrbuf(&(wp -> fixed), _SS_, 1);
+		addstrbuf(&(wp -> fixed), rootpath, 1);
 	}
 
 	if (!w) {
 		if (wp -> path.len <= plen) w++;
-		else if (Xstat(wp -> path.s, &st) < 0) return(argc);
+		else if (stat2(wp -> path.s, &st) < 0) return(argc);
 
 		wp -> s += i;
 		if (isdir) {
@@ -1452,15 +1563,15 @@ wild_t *wp;
 	}
 
 	if (wp -> path.len) cp = wp -> path.s;
-	else if (wp -> fixed.len) cp = _SS_;
-	else cp = ".";
+	else if (wp -> fixed.len) cp = rootpath;
+	else cp = curpath;
 
 	if (!(dirp = Xopendir(cp))) {
 		regexp_free(re);
 		return(argc);
 	}
 	if (wp -> path.len || wp -> fixed.len)
-		addstrbuf(&(wp -> path), _SS_, 1);
+		addstrbuf(&(wp -> path), rootpath, 1);
 
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
@@ -1498,7 +1609,7 @@ wild_t *wp;
 			dupl.ino[(dupl.nino)++].ino = st.st_ino;
 #endif
 
-			addstrbuf(&(dupl.fixed), _SS_, 1);
+			addstrbuf(&(dupl.fixed), rootpath, 1);
 			argc = _evalwild(argc, argvp, &dupl);
 		}
 		else if (regexp_exec(re, dp -> d_name, 1)) {
@@ -1553,6 +1664,7 @@ int flags;
 
 	argv[argc] = NULL;
 	if (argc > 1) qsort(argv, argc, sizeof(char *), cmppath);
+
 	return(argv);
 }
 
@@ -1564,6 +1676,7 @@ char *s;
 	int i;
 
 	for (i = n = 0; s[i]; i++) n += (u_char)(s[i]);
+
 	return(n % MAXHASH);
 }
 
@@ -1589,6 +1702,7 @@ hashlist *next;
 	new -> hits = 0;
 	new -> cost = cost;
 	new -> next = next;
+
 	return(new);
 }
 
@@ -1635,6 +1749,7 @@ hashlist **htable;
 			nextp = &((*nextp) -> next);
 		}
 	}
+
 	return(new);
 }
 
@@ -1646,6 +1761,7 @@ int n;
 
 	for (hp = hashtable[n]; hp; hp = hp -> next)
 		if (!strpathcmp(com, hp -> command)) return(hp);
+
 	return(NULL);
 }
 
@@ -1677,10 +1793,11 @@ int dirok, exe;
 	struct stat st;
 	int d;
 
-	if (Xstat(path, &st) < 0) return(-1);
+	if (stat2(path, &st) < 0) return(-1);
 	d = ((st.st_mode & S_IFMT) == S_IFDIR);
 	if (!exe) return(d);
 	if (d) return(dirok ? d : -1);
+
 	return(Xaccess(path, X_OK));
 }
 
@@ -1705,6 +1822,7 @@ int len, exe;
 		strcpy(&(path[len]), "BAT");
 		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT | CM_BATCH);
 	}
+
 	return(-1);
 }
 #endif	/* MSDOS */
@@ -1717,10 +1835,10 @@ char *com, *search;
 	char *ext;
 #endif
 	char *cp, *tmp, *next, *path;
-	int len, dlen, cost, size, ret, recalc;
+	int len, dlen, cost, size, ret;
 #ifndef	_NOUSEHASH
 	hashlist *hp;
-	int n, duperrno;
+	int n, recalc, duperrno;
 
 	if (!hpp || (!com && !search)) {
 		duperrno = errno;
@@ -1786,7 +1904,9 @@ char *com, *search;
 	}
 #endif	/* CWDINPATH */
 
+#ifndef	_NOUSEHASH
 	recalc = 0;
+#endif
 	if ((next = (search) ? search : getconstvar("PATH"))) {
 		len = strlen(com);
 		size = ret = 0;
@@ -1797,7 +1917,9 @@ char *com, *search;
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 			if (_dospath(cp)) next += 2;
 #endif
+#ifndef	_NOUSEHASH
 			if (*next != _SC_) recalc = CM_RECALC;
+#endif
 			next = strchr(next, PATHDELIM);
 			dlen = (next) ? (next++) - cp : strlen(cp);
 			if (!dlen) tmp = NULL;
@@ -1864,18 +1986,20 @@ char **argv;
 
 	for (i = 0; i < argc; i++)
 		if (!strpathcmp(argv[i], target)) return(argv[i]);
+
 	return(NULL);
 }
 
 # ifndef	NOUID
-static int NEAR completeuser(name, len, argc, argvp)
+int completeuser(name, len, argc, argvp, home)
 char *name;
 int len, argc;
 char ***argvp;
+int home;
 {
 	struct passwd *pwd;
 	char *new;
-	int size;
+	ALLOC_T size;
 
 	len = strlen(name);
 #  ifdef	DEBUG
@@ -1896,10 +2020,13 @@ char ***argvp;
 		}
 		if (!pwd) break;
 		if (strnpathcmp(name, pwd -> pw_name, len)) continue;
-		size = strlen(pwd -> pw_name);
-		new = malloc2(size + 2 + 1);
-		new[0] = '~';
-		strcatdelim2(&(new[1]), pwd -> pw_name, NULL);
+		if (!home) new = strdup2(pwd -> pw_name);
+		else {
+			size = strlen(pwd -> pw_name);
+			new = malloc2(size + 2 + 1);
+			new[0] = '~';
+			strcatdelim2(&(new[1]), pwd -> pw_name, NULL);
+		}
 		if (finddupl(new, argc, *argvp)) {
 			free(new);
 			continue;
@@ -1920,10 +2047,13 @@ char ***argvp;
 	setpwent();
 	while ((pwd = getpwent())) {
 		if (strnpathcmp(name, pwd -> pw_name, len)) continue;
-		size = strlen(pwd -> pw_name);
-		new = malloc2(size + 2 + 1);
-		new[0] = '~';
-		strcatdelim2(&(new[1]), pwd -> pw_name, NULL);
+		if (!home) new = strdup2(pwd -> pw_name);
+		else {
+			size = strlen(pwd -> pw_name);
+			new = malloc2(size + 2 + 1);
+			new[0] = '~';
+			strcatdelim2(&(new[1]), pwd -> pw_name, NULL);
+		}
 		if (finddupl(new, argc, *argvp)) {
 			free(new);
 			continue;
@@ -1935,6 +2065,71 @@ char ***argvp;
 	}
 	endpwent();
 #  endif	/* !DEBUG */
+
+	return(argc);
+}
+
+int completegroup(name, len, argc, argvp)
+char *name;
+int len, argc;
+char ***argvp;
+{
+	struct group *grp;
+	char *new;
+
+	len = strlen(name);
+#  ifdef	DEBUG
+	_mtrace_file = "setgrent(start)";
+	setgrent();
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "setgrent(end)";
+		malloc(0);	/* dummy alloc */
+	}
+	for (;;) {
+		_mtrace_file = "getgrent(start)";
+		grp = getgrent();
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "getgrent(end)";
+			malloc(0);	/* dummy alloc */
+		}
+		if (!grp) break;
+		if (strnpathcmp(name, grp -> gr_name, len)) continue;
+		new = strdup2(grp -> gr_name);
+		if (finddupl(new, argc, *argvp)) {
+			free(new);
+			continue;
+		}
+
+		*argvp = (char **)realloc2(*argvp,
+			(argc + 1) * sizeof(char *));
+		(*argvp)[argc++] = new;
+	}
+	_mtrace_file = "endgrent(start)";
+	endgrent();
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "endgrent(end)";
+		malloc(0);	/* dummy alloc */
+	}
+#  else	/* !DEBUG */
+	setgrent();
+	while ((grp = getgrent())) {
+		if (strnpathcmp(name, grp -> gr_name, len)) continue;
+		new = strdup2(grp -> gr_name);
+		if (finddupl(new, argc, *argvp)) {
+			free(new);
+			continue;
+		}
+
+		*argvp = (char **)realloc2(*argvp,
+			(argc + 1) * sizeof(char *));
+		(*argvp)[argc++] = new;
+	}
+	endgrent();
+#  endif	/* !DEBUG */
+
 	return(argc);
 }
 # endif	/* !NOUID */
@@ -1948,21 +2143,24 @@ int dlen, exe;
 	DIR *dirp;
 	struct dirent *dp;
 	char *cp, *new, path[MAXPATHLEN];
-	int d, size;
+	int d, size, dirok;
 
 	if (dlen >= MAXPATHLEN - 2) return(argc);
 	strncpy2(path, dir, dlen);
 	if (!(dirp = Xopendir(path))) return(argc);
 	cp = strcatdelim(path);
 
+	dirok = (exe <= 1) ? 1 : 0;
 	while ((dp = Xreaddir(dirp))) {
 		if ((!len && isdotdir(dp -> d_name))
-		|| strnpathcmp(file, dp -> d_name, len)) continue;
+		|| strnpathcmp(file, dp -> d_name, len))
+			continue;
 		size = strlen(dp -> d_name);
 		if (size + (cp - path) >= MAXPATHLEN) continue;
 		strncpy2(cp, dp -> d_name, size);
 
-		if ((d = isexecute(path, 1, exe)) < 0) continue;
+		if (isdotdir(dp -> d_name)) d = 1;
+		else if ((d = isexecute(path, dirok, exe)) < 0) continue;
 
 		new = malloc2(size + 1 + 1);
 		strncpy(new, dp -> d_name, size);
@@ -1978,6 +2176,7 @@ int dlen, exe;
 		(*argvp)[argc++] = new;
 	}
 	Xclosedir(dirp);
+
 	return(argc);
 }
 
@@ -1990,7 +2189,7 @@ char ***argvp;
 	int dlen;
 
 # ifdef	CWDINPATH
-	argc = completefile(file, len, argc, argvp, ".", 1, 1);
+	argc = completefile(file, len, argc, argvp, curpath, 1, 2);
 # endif
 	if (!(next = getconstvar("PATH"))) return(argc);
 	for (cp = next; cp; cp = next) {
@@ -2002,9 +2201,10 @@ char ***argvp;
 		dlen = (next) ? (next++) - cp : strlen(cp);
 		tmp = _evalpath(cp, cp + dlen, 0);
 		dlen = strlen(tmp);
-		argc = completefile(file, len, argc, argvp, tmp, dlen, 1);
+		argc = completefile(file, len, argc, argvp, tmp, dlen, 2);
 		free(tmp);
 	}
+
 	return(argc);
 }
 
@@ -2014,6 +2214,9 @@ int len, argc;
 char ***argvp;
 int exe;
 {
+# if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+	char cwd[4];
+# endif
 	char *file, *dir;
 	int dlen;
 
@@ -2033,11 +2236,18 @@ int exe;
 	}
 # ifndef	NOUID
 	else if (*path == '~')
-		return(completeuser(&(path[1]), len - 1, argc, argvp));
+		return(completeuser(&(path[1]), len - 1, argc, argvp, 1));
 # endif
-	else if (exe && dir == path)
-		return(completeexe(path, len, argc, argvp));
-	return(completefile(path, len, argc, argvp, ".", 1, exe));
+# if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+	else if (dir != path) {
+		len -= dlen;
+		dlen = gendospath(cwd, *path, '.') - cwd;
+		return(completefile(dir, len, argc, argvp, cwd, dlen, exe));
+	}
+# endif
+	else if (exe) return(completeexe(path, len, argc, argvp));
+
+	return(completefile(path, len, argc, argvp, curpath, 1, exe));
 }
 
 char *findcommon(argc, argv)
@@ -2059,45 +2269,34 @@ char **argv;
 		free(common);
 		return(NULL);
 	}
+
 	return(common);
 }
 #endif	/* !FDSH && !_NOCOMPLETE */
 
-static int NEAR addmeta(s1, s2, quoted, flags)
+static int NEAR addmeta(s1, s2, quoted)
 char *s1, *s2;
-int quoted, flags;
+int quoted;
 {
-	int i, j;
+	int i, j, pc, quote;
 
 	if (!s2) return(0);
 	for (i = j = 0; s2[i]; i++, j++) {
-		if (s2[i] == '"') {
+		quote = quoted;
+		pc = parsechar(&(s2[i]), -1, '\0', EA_EOLMETA, &quote, NULL);
+		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE || pc == PC_ESCAPE) {
 			if (s1) s1[j] = PMETA;
 			j++;
 		}
-		else if (s2[i] == PMETA) {
-			if (s1) s1[j] = PMETA;
-			j++;
-			if (!(flags & EA_KEEPMETA) && s2[i + 1] == PMETA) i++;
-		}
-		else if (!quoted && s2[i] == '\'') {
-			if (s1) s1[j] = PMETA;
-			j++;
-		}
-		else if (iskanji1(s2, i)) {
+		else if (pc == PC_WORD) {
 			if (s1) s1[j] = s2[i];
 			j++;
 			i++;
 		}
-#ifdef	CODEEUC
-		else if (isekana(s2, i)) {
-			if (s1) s1[j] = s2[i];
-			j++;
-			i++;
-		}
-#endif
+
 		if (s1) s1[j] = s2[i];
 	}
+
 	return(j);
 }
 
@@ -2110,7 +2309,7 @@ int delim;
 
 	if (!argv) return(NULL);
 	for (i = len = 0; argv[i]; i++) len += strlen(argv[i]);
-	if (i < 1) return(strdup2(""));
+	if (i < 1) return(strdup2(nullstr));
 	len += (delim) ? i - 1 : 0;
 	cp = malloc2(len + 1);
 	len = strcpy2(cp, argv[0]) - cp;
@@ -2118,6 +2317,7 @@ int delim;
 		if (delim) cp[len++] = delim;
 		len = strcpy2(&(cp[len]), argv[i]) - cp;
 	}
+
 	return(cp);
 }
 
@@ -2128,6 +2328,7 @@ char **var;
 
 	if (!var) return(0);
 	for (i = 0; var[i]; i++);
+
 	return(i);
 }
 
@@ -2159,6 +2360,7 @@ int margin;
 	dupl = (char **)malloc2((n + margin + 1) * sizeof(char *));
 	for (i = 0; i < n; i++) dupl[i] = strdup2(var[i]);
 	dupl[i] = NULL;
+
 	return(dupl);
 }
 
@@ -2167,6 +2369,10 @@ char *s;
 int len, spc, flags, *qp, *pqp;
 {
 	if (*s == *qp) {
+#ifdef	FD
+		if (flags & EA_FINDMETA)
+			return((*qp == '\'') ? PC_EXMETA : PC_META);
+#endif
 		if (!pqp) *qp = '\0';
 		else {
 			*qp = *pqp;
@@ -2177,25 +2383,29 @@ int len, spc, flags, *qp, *pqp;
 	else if (iskanji1(s, 0)) return(PC_WORD);
 #ifdef	CODEEUC
 	else if (isekana(s, 0)) return(PC_WORD);
-#endif
-#ifdef	BASHSTYLE
-	else if (*qp == '`') return(PC_BQUOTE);
 #endif
 	else if (*qp == '\'') return(PC_SQUOTE);
-	else if (spc && *s == spc) return(*s);
-	else if (ismeta(s, 0, *qp, len)) return(PC_META);
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
+#ifdef	FD
+	else if ((flags & EA_FINDMETA) && strchr(DQ_METACHAR, *s))
+		return(PC_META);
+	else if ((flags & EA_FINDDELIM) && strchr(CMDLINE_DELIM, *s))
+		return(PC_DELIM);
+#endif
+	else if (isescape(s, 0, *qp, len, flags)) return(PC_ESCAPE);
+	else if (*qp == '`') return(PC_BQUOTE);
+#ifdef	NESTINGQUOTE
 	else if ((flags & EA_BACKQ) && *s == '`') {
 		if (pqp && *qp) *pqp = *qp;
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
-#else
-	else if (*qp == '`') return(PC_BQUOTE);
 #endif
+	else if (spc && *s == spc) return(*s);
 	else if (*qp) return(PC_DQUOTE);
 	else if (!(flags & EA_NOEVALQ) && *s == '\'') {
+#ifdef	FD
+		if (flags & EA_FINDMETA) return(PC_META);
+#endif
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
@@ -2203,14 +2413,17 @@ int len, spc, flags, *qp, *pqp;
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
-#ifndef	BASHSTYLE
-	/* bash can include `...` in "..." */
+#ifndef	NESTINGQUOTE
 	else if ((flags & EA_BACKQ) && *s == '`') {
 		if (pqp && *qp) *pqp = *qp;
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
 #endif
+#ifdef	FD
+	else if ((flags & EA_FINDMETA) && strchr(METACHAR, *s))
+		return(PC_META);
+#endif
 
 	return(PC_NORMAL);
 }
@@ -2302,6 +2515,7 @@ int *eolp, *ptrp, qed;
 	if (skipvarvalue(*bufp, ptrp, "}", qed, 1, 0) < 0) return(-1);
 # endif
 #endif	/* BASHSTYLE */
+
 	return(mode);
 }
 
@@ -2319,25 +2533,24 @@ char *next;
 int qed, nonl, nest;
 #endif
 {
-#ifdef	BASHSTYLE
+#ifdef	NESTINGQUOTE
 	int pq;
 #endif
 	char *cp;
 	int pc, q, len;
 
-#ifdef	BASHSTYLE
+#ifdef	NESTINGQUOTE
 	pq = '\0';
 #endif
 	q = '\0';
 	len = strlen(next);
 	while (s[*ptrp]) {
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
+#ifdef	NESTINGQUOTE
 		pc = parsechar(&(s[*ptrp]), -1, '$', EA_BACKQ, &q, &pq);
 #else
 		pc = parsechar(&(s[*ptrp]), -1, '$', EA_BACKQ, &q, NULL);
 #endif
-		if (pc == PC_WORD || pc == PC_META) (*ptrp)++;
+		if (pc == PC_WORD || pc == PC_ESCAPE) (*ptrp)++;
 		else if (pc == '$') {
 			if (!s[++(*ptrp)]) return(0);
 			cp = s;
@@ -2366,6 +2579,7 @@ int qed, nonl, nest;
 
 		(*ptrp)++;
 	}
+
 	return(-1);
 }
 
@@ -2425,6 +2639,7 @@ int plen, mode;
 		free(tmp);
 	}
 	regexp_free(re);
+
 	return(ret);
 }
 
@@ -2443,6 +2658,7 @@ int plen, mode;
 		else new[i] = strdup2(var[i]);
 	}
 	new[i] = NULL;
+
 	return(new);
 }
 #endif	/* !MINIMUMSHELL */
@@ -2491,20 +2707,18 @@ int plen, *modep;
 #endif
 
 			for (i = j = 0; arglist[i + 1]; i++)
-				j += addmeta(NULL,
-					arglist[i + 1], quoted, EA_KEEPMETA);
-			if (i <= 0) cp = strdup2("");
+				j += addmeta(NULL, arglist[i + 1], quoted);
+			if (i <= 0) cp = strdup2(nullstr);
 			else {
 				j += (i - 1) * 3;
 				cp = malloc2(j + 1);
-				j = addmeta(cp, arglist[1],
-					quoted, EA_KEEPMETA);
+				j = addmeta(cp, arglist[1], quoted);
 				for (i = 2; arglist[i]; i++) {
 					cp[j++] = quoted;
 					cp[j++] = sp;
 					cp[j++] = quoted;
 					j += addmeta(&(cp[j]), arglist[i],
-						quoted, EA_KEEPMETA);
+						quoted);
 				}
 				cp[j] = '\0';
 			}
@@ -2561,6 +2775,7 @@ int plen, *modep;
 #ifndef	MINIMUMSHELL
 	freevar(new);
 #endif
+
 	return(cp);
 }
 
@@ -2619,6 +2834,7 @@ int s, len, vlen, mode, quoted;
 		if (exitfunc) (*exitfunc)();
 		return(-2);
 	}
+
 	return(mode);
 }
 
@@ -2630,6 +2846,7 @@ int olen, nlen;
 {
 	if (nlen <= olen) return(buf);
 	buf = realloc2(buf, ptr + nlen + (int)strlen(arg) - olen + 1);
+
 	return(buf);
 }
 
@@ -2746,9 +2963,9 @@ int quoted;
 	}
 
 	if (!mode && (c != '@' || !quoted)) {
-		vlen = addmeta(NULL, cp, quoted, EA_KEEPMETA);
+		vlen = addmeta(NULL, cp, quoted);
 		*bufp = insertarg(*bufp, ptr, arg, *argp - arg + 1, vlen);
-		addmeta(&((*bufp)[ptr]), cp, quoted, EA_KEEPMETA);
+		addmeta(&((*bufp)[ptr]), cp, quoted);
 	}
 	else if (!cp) vlen = 0;
 	else if (c == '@' && !*cp && quoted
@@ -2765,6 +2982,7 @@ int quoted;
 
 	ptr += vlen;
 	if (new) free(new);
+
 	return(ptr);
 }
 
@@ -2813,6 +3031,7 @@ char *name;
 	uidlist[maxuid].uid = pwd -> pw_uid;
 	uidlist[maxuid].name = strdup2(pwd -> pw_name);
 	uidlist[maxuid].home = strdup2(pwd -> pw_dir);
+
 	return(&(uidlist[maxuid++]));
 }
 
@@ -2861,6 +3080,7 @@ char *name;
 	gidlist[maxgid].name = strdup2(grp -> gr_name);
 	gidlist[maxgid].gr_mem = duplvar(grp -> gr_mem, -1);
 	gidlist[maxgid].ismem = 0;
+
 	return(&(gidlist[maxgid++]));
 }
 
@@ -2923,12 +3143,12 @@ int rest;
 	char *tmp;
 	int len, size;
 
-	stripquote(bbuf, 0);
+	stripquote(bbuf, EA_BACKQ);
 	if (!(tmp = (*backquotefunc)(bbuf))) return(buf);
-	len = addmeta(NULL, tmp, '\0', 0);
+	len = addmeta(NULL, tmp, '\0');
 	size = *ptrp + len + rest + 1;
 	buf = realloc2(buf, size);
-	addmeta(&(buf[*ptrp]), tmp, '\0', 0);
+	addmeta(&(buf[*ptrp]), tmp, '\0');
 	*ptrp += len;
 	free(tmp);
 
@@ -2965,6 +3185,7 @@ char *gethomedir(VOID_A)
 	if (pwd) return(pwd -> pw_dir);
 # endif	/* !FD */
 #endif	/* !NOUID */
+
 	return(NULL);
 }
 
@@ -2982,7 +3203,7 @@ char **argp;
 	len = ((cp = strdelim(top, 0))) ? (cp - top) : strlen(top);
 	if (!len) cp = gethomedir();
 # ifdef	FD
-	else if (len == sizeof("FD") - 1 && !strnpathcmp(top, "FD", len))
+	else if (len == strsize("FD") && !strnpathcmp(top, "FD", len))
 		cp = progpath;
 # endif
 	else {
@@ -3029,6 +3250,7 @@ char **argp;
 	strncpy(&((*bufp)[ptr]), cp, vlen);
 	ptr += vlen;
 	*argp += len;
+
 	return(ptr);
 }
 #endif	/* !MINIMUMSHELL */
@@ -3040,11 +3262,14 @@ int qed, flags;
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char path[MAXPATHLEN], alias[MAXPATHLEN];
 #endif
-#ifdef	BASHSTYLE
+#ifndef	MINIMUMSHELL
+	int prev;
+#endif
+#ifdef	NESTINGQUOTE
 	int pq;
 #endif
 	char *cp, *buf, *bbuf;
-	int i, j, pc, prev, q;
+	int i, j, pc, q;
 
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	if (*arg == '"' && (i = strlen(arg)) > 2 && arg[i - 1] == '"') {
@@ -3067,15 +3292,18 @@ int qed, flags;
 	buf = malloc2(i);
 	if (!backquotefunc) flags &= ~EA_BACKQ;
 	bbuf = (flags & EA_BACKQ) ? malloc2(i) : NULL;
-#ifdef	BASHSTYLE
-	pq = '\0';
+#ifndef	MINIMUMSHELL
+	prev =
 #endif
-	prev = q = '\0';
+#ifdef	NESTINGQUOTE
+	pq =
+#endif
+	q = '\0';
 	i = j = 0;
+	cp = arg;
 
-	for (cp = arg; *cp; prev = *(cp++)) {
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
+	while (*cp) {
+#ifdef	NESTINGQUOTE
 		pc = parsechar(cp, -1, '$', flags, &q, &pq);
 #else
 		pc = parsechar(cp, -1, '$', flags, &q, NULL);
@@ -3112,7 +3340,7 @@ int qed, flags;
 # else
 			else if (cp[1] == '$' || cp[1] == '~') cp++;
 # endif
-#endif
+#endif	/* FAKEMETA */
 			else if ((i = evalvar(&buf, i, &cp, tmpq)) < 0) {
 				if (bbuf) free(bbuf);
 				free(buf);
@@ -3120,16 +3348,24 @@ int qed, flags;
 				return(NULL);
 			}
 		}
-		else if (pc == PC_META) {
+		else if (pc == PC_ESCAPE) {
 			cp++;
-			if (flags & EA_KEEPMETA) buf[i++] = PMETA;
+			if (flags & EA_KEEPMETA) pc = PC_NORMAL;
 			else if (*cp == '$') /*EMPTY*/;
 			else if ((flags & EA_BACKQ) && *cp == '`') /*EMPTY*/;
 			else if ((flags & EA_STRIPQ)
 			&& (*cp == '\'' || *cp == '"'))
 				/*EMPTY*/;
-			else buf[i++] = PMETA;
-			buf[i++] = *cp;
+			else pc = PC_NORMAL;
+
+			if (q == '`') {
+				bbuf[j++] = PMETA;
+				bbuf[j++] = *cp;
+			}
+			else {
+				if (pc != PC_ESCAPE) buf[i++] = PMETA;
+				buf[i++] = *cp;
+			}
 		}
 		else if (pc == PC_OPQUOTE) {
 			if (*cp == '`') j = 0;
@@ -3141,6 +3377,12 @@ int qed, flags;
 			i = evalhome(&buf, i, &cp);
 #endif
 		else buf[i++] = *cp;
+
+#ifdef	MINIMUMSHELL
+		cp++;
+#else
+		prev = *(cp++);
+#endif
 	}
 #ifndef	BASHSTYLE
 	/* bash does not allow unclosed quote */
@@ -3168,7 +3410,7 @@ char ***argvp, *ifs;
 		for (i = 0, quote = '\0'; (*argvp)[n][i]; i++) {
 			pc = parsechar(&((*argvp)[n][i]), -1,
 				'\0', 0, &quote, NULL);
-			if (pc == PC_WORD || pc == PC_META) i++;
+			if (pc == PC_WORD || pc == PC_ESCAPE) i++;
 			else if (pc != PC_NORMAL) /*EMPTY*/;
 			else if (strchr2(ifs, (*argvp)[n][i])) {
 				for (j = i + 1; (*argvp)[n][j]; j++)
@@ -3201,6 +3443,7 @@ char ***argvp, *ifs;
 				(argc-- - n) * sizeof(char *));
 		}
 	}
+
 	return(argc);
 }
 
@@ -3233,6 +3476,7 @@ int flags;
 		free(wild);
 		n += i - 1;
 	}
+
 	return(argc);
 }
 
@@ -3247,22 +3491,24 @@ int flags;
 	for (i = j = 0, quote = '\0'; arg[i]; i++) {
 		pc = parsechar(&(arg[i]), -1, '\0', 0, &quote, NULL);
 		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) {
-			stripped = 1;
+			stripped++;
 			if (flags & EA_STRIPQ) continue;
 		}
 		else if (pc == PC_WORD) arg[j++] = arg[i++];
-		else if (pc == PC_META) {
-			stripped = 1;
-			if ((flags & EA_KEEPMETA)
-			|| quote == '\''
-			|| (quote == '"' && !strchr(DQ_METACHAR, arg[i + 1])))
-				arg[j++] = arg[i];
+		else if (pc == PC_ESCAPE) {
 			i++;
+			if (flags & EA_KEEPMETA) pc = PC_NORMAL;
+			else if (!quote && !(flags & EA_BACKQ)) /*EMPTY*/;
+			else if (!strchr(DQ_METACHAR, arg[i])) pc = PC_NORMAL;
+
+			if (pc != PC_ESCAPE) arg[j++] = PMETA;
+			else stripped++;
 		}
 
 		arg[j++] = arg[i];
 	}
 	arg[j] = '\0';
+
 	return(stripped);
 }
 
@@ -3312,11 +3558,10 @@ int flags;
 			if (!(flags & EA_NOEVALQ)) continue;
 		}
 		else if (pc == PC_WORD) tmp[j++] = cp[i++];
-		else if (pc == PC_META) {
+		else if (pc == PC_ESCAPE) {
 			i++;
 			if ((flags & EA_KEEPMETA)
-			|| quote == '\''
-			|| (quote == '"' && !strchr(DQ_METACHAR, cp[i])))
+			|| (quote && !strchr(DQ_METACHAR, cp[i])))
 				tmp[j++] = PMETA;
 		}
 		else if (pc == PC_OPQUOTE) {
@@ -3333,6 +3578,7 @@ int flags;
 	}
 	tmp[j] = '\0';
 	free(cp);
+
 	return(tmp);
 }
 
@@ -3346,5 +3592,6 @@ int flags;
 	for (cp = path; *cp == ' ' || *cp == '\t'; cp++);
 	cp = _evalpath(cp, NULL, flags);
 	free(path);
+
 	return(cp);
 }
